# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}
from importlib.metadata import distribution, PackageNotFoundError

try:
    distribution('volttron-core')
    from volttron.client.messaging import topics as t
except PackageNotFoundError:
    from volttron.platform.messaging import topics as t

VALUE_RESPONSE_PREFIX = t.ACTUATOR_VALUE()
REVERT_POINT_RESPONSE_PREFIX = t.ACTUATOR_REVERTED_POINT()
REVERT_DEVICE_RESPONSE_PREFIX = t.ACTUATOR_REVERTED_DEVICE()
ERROR_RESPONSE_PREFIX = t.ACTUATOR_ERROR()

WRITE_ATTEMPT_PREFIX = t.ACTUATOR_WRITE()

GET_TOPIC = t.ACTUATOR_GET()
RESERVATION_REQUEST_TOPIC = t.ACTUATOR_SCHEDULE_REQUEST()
RESERVATION_RESULT_TOPIC = t.ACTUATOR_SCHEDULE_RESULT()
REVERT_DEVICE_TOPIC = t.ACTUATOR_REVERT_DEVICE()
REVERT_POINT_TOPIC = t.ACTUATOR_REVERT_POINT()
SET_TOPIC = t.ACTUATOR_SET()

RESERVATION_ACTION_NEW = 'NEW_RESERVATION'
RESERVATION_ACTION_CANCEL = 'CANCEL_RESERVATION'
LEGACY_RESERVATION_ACTION_NEW = 'NEW_SCHEDULE'
LEGACY_RESERVATION_ACTION_CANCEL = 'CANCEL_SCHEDULE'

RESERVATION_RESPONSE_SUCCESS = 'SUCCESS'
RESERVATION_RESPONSE_FAILURE = 'FAILURE'

RESERVATION_CANCEL_PREEMPTED = 'PREEMPTED'

ACTUATOR_COLLECTION = 'actuators'


# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

import gevent
import logging
import re
import subprocess
import sys

from collections import defaultdict
from datetime import datetime
from importlib.metadata import distribution, PackageNotFoundError
from pathlib import Path
from pkgutil import iter_modules
from pydantic import ValidationError
from typing import Any, Iterable, Sequence, Set

try:
    import tomllib  # Will not be available below 3.11. Need to pip install tomli.
except ModuleNotFoundError:
    import tomli as tomllib

try:
    distribution('volttron-core')
    from argparse import Namespace
    from contextlib import redirect_stdout
    from io import StringIO
    from volttron.client.commands.install_parser import install_lib_vctl
    from volttron.client.known_identities import PLATFORM_DRIVER
    from volttron.client.logs import setup_logging
    from volttron.client.messaging.health import STATUS_BAD
    from volttron.client.messaging.utils import normtopic
    from volttron.client.vip.agent import Agent
    from volttron.client.vip.agent.subsystems.rpc import RPC
    from volttron.driver.base.driver import BaseInterface, DriverAgent
    from volttron.driver.base.driver_locks import configure_publish_lock, setup_socket_lock
    from volttron.driver.base.config import DeviceConfig, EquipmentConfig, PointConfig, RemoteConfig
    from volttron.driver.base.utils import publication_headers, publish_wrapper
    from volttron.utils import format_timestamp, get_aware_utc_now, load_config, vip_main
    from volttron.utils.jsonrpc import RemoteError
    from volttron.utils.scheduling import periodic
except PackageNotFoundError:
    from importlib.metadata import requires
    from volttron.platform.agent.known_identities import PLATFORM_DRIVER
    from volttron.platform.messaging.health import STATUS_BAD
    from volttron.platform.messaging.utils import normtopic
    from volttron.platform.vip.agent import Agent
    from volttron.platform.vip.agent.subsystems.rpc import RPC
    from volttron.driver.base.driver import BaseInterface, DriverAgent
    from volttron.driver.base.driver_locks import configure_publish_lock, setup_socket_lock
    from volttron.driver.base.config import DeviceConfig, EquipmentConfig, PointConfig, RemoteConfig
    from volttron.driver.base.utils import publication_headers, publish_wrapper
    from volttron.platform.agent.utils import format_timestamp, get_aware_utc_now, load_config, setup_logging, vip_main
    from volttron.platform.jsonrpc import RemoteError
    from volttron.platform.scheduling import periodic

from .config import PlatformDriverConfig
from .constants import *
from .equipment import DeviceNode, EquipmentNode, EquipmentTree, PointNode
from .overrides import OverrideManager
from .poll_scheduler import PollScheduler
from .reservations import ReservationManager
from .scalability_testing import ScalabilityTester

try:
    distribution('volttron-core')
    from volttron.utils.context import ClientContext as Cc
    logging.basicConfig(filename=f"{Cc.get_volttron_home()}/driver.log", level=logging.DEBUG, format='%(asctime)s %(levelname)s %(name)s %(message)s')
except PackageNotFoundError:
    setup_logging()
_log = logging.getLogger(__name__)
__version__ = '4.0'


class PlatformDriverAgent(Agent):

    def __init__(self, **kwargs):
        config_path = kwargs.pop('config_path', None)
        super(PlatformDriverAgent, self).__init__(**kwargs)
        self.config: PlatformDriverConfig = self._load_agent_config(load_config(config_path) if config_path else {})
        # Initialize internal data structures:
        self.equipment_tree = EquipmentTree(self)
        self.interface_classes = {}

        # Set up locations for helper objects:
        self.heartbeat_greenlet = None
        self.override_manager = None  # TODO: Should this initialize object here and call a load method on config?
        self.poll_schedulers = {}
        self.publishers = {}
        self.reservation_manager = None  # TODO: Should this use a default reservation manager?
        self.scalability_test = None

        self.vip.config.set_default("config", self.config.model_dump())
        self.vip.config.subscribe(self.configure_main, actions=['NEW', 'UPDATE', 'DELETE'], pattern='config')
        self.vip.config.subscribe(self._configure_new_equipment, actions=['NEW'], pattern='devices/*')
        self.vip.config.subscribe(self._update_equipment, actions=['UPDATE'], pattern='devices/*')
        self.vip.config.subscribe(self._remove_equipment, actions='DELETE', pattern='devices/*')

    #########################
    # Configuration & Startup
    #########################

    def _load_agent_config(self, config: dict) -> PlatformDriverConfig:
        try:
            return PlatformDriverConfig(**config)
        except ValidationError as e:
            _log.warning(f'Validation of platform driver configuration file failed. Using default values. --- {str(e)}')
            if self.core.connected:  # TODO: Is this a valid way to make sure we are ready to call subsystems?
                self.vip.health.set_status(STATUS_BAD, f'Error processing configuration: {e}')
            return PlatformDriverConfig()

    def configure_main(self, _, action: str, contents: dict):
        old_config = self.config.model_copy(deep=True)
        new_config = self._load_agent_config(contents)
        if action == "NEW":
            self.config = new_config
            self.equipment_tree = EquipmentTree(self)
            try:
                setup_socket_lock(self.config.max_open_sockets)
                configure_publish_lock(int(self.config.max_concurrent_publishes))
                self.scalability_test = (ScalabilityTester(self.config.scalability_test_iterations)
                                         if self.config.scalability_test else None)
            except ValueError as e:
                _log.error(f"ERROR PROCESSING STARTUP CRITICAL CONFIGURATION SETTINGS: {e}")
                _log.error("Platform driver SHUTTING DOWN")
                sys.exit(1)

        else:
            # Some settings cannot be changed while running. Warn and replace these with the old ones until restart.
            _log.info('Updated configuration received for Platform Driver.')
            if new_config.max_open_sockets != old_config['max_open_sockets']:
                new_config.max_open_sockets = old_config['max_open_sockets']
                _log.info('Restart Platform Driver for changes to the max_open_sockets setting to take effect')

            if new_config.max_concurrent_publishes != old_config['max_concurrent_publishes']:
                new_config.max_concurrent_publishes = old_config['max_concurrent_publishes']
                _log.info('Restart Platform Driver for changes to the max_concurrent_publishes setting to take effect')

            if new_config.scalability_test != old_config['scalability_test']:
                new_config.scalability_test = old_config['scalability_test']
                if not old_config.scalability_test:
                    _log.info('Restart Platform Driver with scalability_test set to true in order to run a test.')
                if old_config.scalability_test:
                    _log.info("A scalability test may not be interrupted. Restart the driver to stop the test.")
            try:
                if new_config.scalability_test_iterations != old_config['scalability_test_iterations'] and \
                        old_config.scalability_test:
                    new_config.scalability_test_iterations = old_config['scalability_test_iterations']
                    _log.info('The scalability_test_iterations setting cannot be changed without restarting the agent.')
            except ValueError:
                pass
            if old_config.scalability_test:
                _log.info("Running scalability test. Settings may not be changed without restart.")
                return
            self.config = new_config
        if self.override_manager is None:
            self.override_manager = OverrideManager(self)

        # Set up Reservation Manager:
        if self.reservation_manager is None:
            now = get_aware_utc_now()
            self.reservation_manager = ReservationManager(self, self.config.reservation_preempt_grace_time, now)
            self.reservation_manager.update(now)
        else:
            self.reservation_manager.set_grace_period(self.config.reservation_preempt_grace_time)

        # Set up heartbeat to devices:
        # TODO: Should this be globally uniform (here), by device (in remote), or globally scheduled (in poll scheduler)?
        # Only restart the heartbeat if it changes.
        if (self.config.remote_heartbeat_interval != old_config.remote_heartbeat_interval
                or action == "NEW" or self.heartbeat_greenlet is None):
            if self.heartbeat_greenlet is not None:
                self.heartbeat_greenlet.kill()
            self.heartbeat_greenlet = self.core.periodic(self.config.remote_heartbeat_interval, self.heart_beat)

        # Start subscriptions:
        current_subscriptions = {topic: subscribed for _, topic, subscribed in self.vip.pubsub.list('pubsub').get()}
        for topic, callback in [
            (GET_TOPIC, self.handle_get),
            (SET_TOPIC, self.handle_set),
            (RESERVATION_REQUEST_TOPIC, self.handle_reservation_request),
            (REVERT_POINT_TOPIC, self.handle_revert_point),
            (REVERT_DEVICE_TOPIC, self.handle_revert_device)
        ]:
            if not current_subscriptions.get(topic):
                self.vip.pubsub.subscribe('pubsub', topic, callback)

        # Load Equipment Tree:
        for c in self.vip.config.list():
            if 'devices/' in c[:8]:
                equipment_config = self.vip.config.get(c)
                self._configure_new_equipment(c, 'NEW', equipment_config, schedule_now=False)

        # Schedule Polling
        self.poll_schedulers = PollScheduler.setup(self.equipment_tree, self.config.groups)
        for poll_scheduler in self.poll_schedulers.values():
            poll_scheduler.schedule()

        # Set up All Publishes:
        self._start_all_publishes()

    def _separate_equipment_configs(self, config_dict: dict[str, Any]
                                    ) -> tuple[RemoteConfig, DeviceConfig | None, list[dict[str, Any]]]:
        # Separate remote_config and make adjustments for possible config version 1:
        remote_config_dict: dict[str, Any] = config_dict.pop('remote_config', config_dict.pop('driver_config', {}))
        remote_config_dict['driver_type'] = remote_config_dict.get('driver_type', config_dict.pop('driver_type', None))
        remote_config_dict['heart_beat_point'] = remote_config_dict.get('heart_beat_point',
                                                              config_dict.pop('heart_beat_point', None))
        remote_config: RemoteConfig = RemoteConfig(**remote_config_dict)
        if remote_config.driver_type:
            # Received new device node.
            registry_config = config_dict.pop('registry_config', [])
            registry_config = registry_config if registry_config is not None else []
            dev_config = DeviceConfig(**config_dict)

            point_configs = []
            # Set up any point nodes which are children of this device.
            for reg in registry_config:
                # If there are fields in device config for all registries, add them where they are not overridden:
                for k, v in dev_config.equipment_specific_fields.items():
                    if not reg.get(k):
                        reg[k] = v
                point_configs.append(reg)

        else:
            dev_config, point_configs = None, []
        return remote_config, dev_config, point_configs

    def _configure_new_equipment(self, equipment_name: str, _, contents: dict, schedule_now: bool = True) -> bool:
        existing_node = self.equipment_tree.get_node(equipment_name)
        if existing_node:
            if not existing_node.config_finished:
                existing_node.config_finished = True
                return False
            else:
                return self._update_equipment(equipment_name, 'UPDATE', contents)
        try:
            remote_config, dev_config, registry_configs = self._separate_equipment_configs(contents)
            if dev_config:
                # Received new device node.
                remote = self._get_or_create_remote(equipment_name, remote_config, dev_config.allow_duplicate_remotes)
                validated_reg_configs = (remote.interface.REGISTER_CONFIG_CLASS(**r) for r in registry_configs)
                device_node = self.equipment_tree.add_device(device_topic=equipment_name, dev_config=dev_config,
                                                             remote=remote, registry_configs=validated_reg_configs)
                remote.add_equipment(device_node)
            else: # Received new or updated segment node.
                equipment_config = EquipmentConfig(**contents)
                self.equipment_tree.add_segment(equipment_name, equipment_config)
            if schedule_now:
                points = self.equipment_tree.points(equipment_name)
                self._update_polling_schedules(points)
            return True
        except ValueError as e:
            _log.warning(f'Skipping configuration of equipment: {equipment_name} after encountering error --- {e}')
            return False

    def _get_or_create_remote(self, equipment_name: str, remote_config: RemoteConfig, allow_duplicate_remotes: bool,
                              is_update: bool = False):
        interface = self._get_configured_interface(remote_config)
        allow_duplicate_remotes = True if (allow_duplicate_remotes or self.config.allow_duplicate_remotes) else False
        if not allow_duplicate_remotes:
            unique_remote_id = interface.unique_remote_id(equipment_name, remote_config)
        else:
            unique_remote_id = BaseInterface.unique_remote_id(equipment_name, remote_config)

        remote = self.equipment_tree.remotes.get(unique_remote_id)
        if not remote:
            remote = DriverAgent(remote_config, self.core, self.equipment_tree, self.scalability_test,
                                       self.config.timezone, unique_remote_id, self.vip)
            self.equipment_tree.remotes[unique_remote_id] = remote
        elif not is_update and remote.config != remote.interface.INTERFACE_CONFIG_CLASS(**remote_config.model_dump()):
            # TODO: Can we support some settings being different between two groupings on same remote?
            #       e.g., two groups with different cov_lifetime intervals?
            #       (This doesn't affect remote itself, but is an interface specific configuration.)
            #       Can this use case already be accomodated using groups somehow?
            _log.warning(f'Remote configuration for equipment "{equipment_name}" does not match configuration'
                         f' of shared remote "{unique_remote_id}. Check configurations for consistency or consider'
                         f' setting "allow_duplicate_remotes == True".')
        return remote

    def _get_configured_interface(self, remote_config: RemoteConfig):
        driver_type = 'fake' if remote_config.driver_type == 'fakedriver' else remote_config.driver_type
        interface = self.interface_classes.get(driver_type)
        if not interface:
            try:
                module = remote_config.module
                interface = BaseInterface.get_interface_subclass(driver_type, module)
                if interface.default_config is None:
                    try:
                        interface.default_config = self.vip.config.get(f'interfaces/{driver_type}')
                    except KeyError:  # TODO: Can this even raise KeyError? It will be None or and Attribute Error, no?
                        interface.default_config = {}
            except (AttributeError, ModuleNotFoundError, ValueError) as e:
                raise ValueError(f'Unable to configure driver with interface: {driver_type}.'
                                 f' This interface type is currently unknown or not installed.'
                                 f' Received exception: {e}')
            self.interface_classes[driver_type] = interface
        return interface

    def _update_equipment(self, config_name: str, _, contents: dict) -> bool:
        """Callback for updating equipment configuration."""
        remote_config, dev_config, registry_configs = self._separate_equipment_configs(contents)
        if dev_config:
            try:
                remote = self._get_or_create_remote(config_name, remote_config, dev_config.allow_duplicate_remotes, True)
            except ValueError as e:
                _log.warning(f'Skipping configuration of equipment: {config_name} after encountering error --- {e}')
                return False
            if remote.config != remote.interface.INTERFACE_CONFIG_CLASS(**remote_config.model_dump()):
                # TODO: How do we reconcile potential differences with other users of this remote?
                #       Currently, we are just going to change things out from underneath them.
                pass
        else:
            remote = None
        validated_reg_configs = [remote.interface.REGISTER_CONFIG_CLASS(**r) for r in registry_configs]
        is_changed = self.equipment_tree.update_equipment(config_name, dev_config, remote, validated_reg_configs)
        if is_changed:
            points = self.equipment_tree.points(config_name)
            self._update_polling_schedules(points)
        return is_changed

    def _update_polling_schedules(self, points):
        reschedules_required, new_groups = [], []
        for point in points:
            group = self.equipment_tree.get_group(point.identifier)
            if group not in self.poll_schedulers:
                new_groups.append(group)
            if PollScheduler.add_to_schedule(point, self.equipment_tree):
                reschedules_required.append(group)
        if new_groups:
            self.poll_schedulers.update(PollScheduler.create_poll_schedulers(self.equipment_tree, self.config.groups,
                                                                             new_groups, len(self.poll_schedulers)))
        for updated_group in reschedules_required:
            self.poll_schedulers[updated_group].schedule()

    def _remove_equipment(self, config_name: str, _, __, leave_disconnected=False) -> bool:
        """Callback to remove equipment configuration."""
        poll_schedulers = []
        for point in self.equipment_tree.points(config_name):
            group = self.equipment_tree.get_group(point.identifier)
            poll_schedulers.append(self.poll_schedulers.get(group))
        removed_count = self.equipment_tree.remove_segment(config_name, leave_disconnected)
        # TODO: Add reschedule_all_on_update option and reschedule all poll_schedulers when true.
        return True if removed_count > 0 else False

    def _start_all_publishes(self):
        # TODO: Can we just schedule and let the stale property work its magic?
        for device in self.equipment_tree.devices(self.equipment_tree.root):
            if (device.all_publish_interval and
                    (self.equipment_tree.is_published_all_depth(device.identifier) or
                     self.equipment_tree.is_published_all_breadth(device.identifier))):
                # Schedule first publish at end of first polling cycle to guarantee all points should have data.
                start_all_datatime = max(poller.start_all_datetime for poller in self.poll_schedulers.values())
                self.publishers[device] = self.core.schedule(
                    periodic(device.all_publish_interval, start=start_all_datatime), self._all_publish, device
                )

    def _all_publish(self, node):
        if not node:
            _log.warning(f'All publish running for a device node which no longer exists.')
        active_points = self.equipment_tree.active_points(node)
        if not ((ready_points := self.equipment_tree.ready_points(node))
                or (self.equipment_tree.strict_all_publishes(node.identifier))
                    and active_points != ready_points):
            _log.info(f'Skipping all publish of device: {node.identifier}. Data is not yet ready.')
            return
        if not ((non_stale_points := self.equipment_tree.non_stale_points(ready_points))
                or (self.equipment_tree.strict_all_publishes(node.identifier))
                    and active_points != non_stale_points):
            _log.warning(f'Skipping all publish of device: {node.identifier}. Data is stale.')
            return
        else:
            headers = publication_headers()
            depth_topic, breadth_topic = self.equipment_tree.get_device_topics(node.identifier)
            if self.equipment_tree.is_published_all_depth(node.identifier):
                publish_wrapper(self.vip, f'{depth_topic}/all', headers=headers, message=[
                    {p.identifier.rsplit('/', 1)[-1]: p.last_value for p in non_stale_points},
                    {p.identifier.rsplit('/', 1)[-1]: p.meta_data for p in non_stale_points}
                ])
            elif self.equipment_tree.is_published_all_breadth(node.identifier):
                publish_wrapper(self.vip, f'{breadth_topic}/all', headers=headers, message=[
                    {p.identifier.rsplit('/', 1)[-1]: p.last_value for p in non_stale_points},
                    {p.identifier.rsplit('/', 1)[-1]: p.meta_data for p in non_stale_points}
                ])

    ###############
    # Query Backend
    ###############

    def semantic_query(self, query):
        """ Resolve tags from tagging service. """
        try:
            return self.vip.rpc.call('platform.semantic', 'semantic_query', query).get(timeout=5)
        except gevent.Timeout as e:
            _log.warning(f'Semantic Interoperability Service timed out: {e.exception}')
            return {}

    def build_query_plan(self, topic: str | Sequence[str] | Set[str] = None,
                         regex: str = None) -> dict[DriverAgent, Set[PointNode]]:
        """ Find points to be queried and organize by remote."""
        exact_matches, topic = (topic, None) if isinstance(topic, list) or isinstance(topic, set) else ([], topic)
        query_plan = defaultdict(set)
        for p in self.equipment_tree.find_points(topic, regex, exact_matches):
            query_plan[self.equipment_tree.get_remote(p.identifier)].add(p)
        return query_plan

    ###############
    # RPC Interface
    ###############

    @RPC.export
    def get(self, topic: str | Sequence[str] | Set[str] = None, regex: str = None) -> tuple[dict, dict]:
        # Find set of points to query and organize by remote:
        query_plan = self.build_query_plan(topic, regex)
        return self._get(query_plan)

    @RPC.export
    def semantic_get(self, query: str) -> tuple[dict, dict]:
        exact_matches = self.semantic_query(query)
        query_plan = self.build_query_plan(exact_matches)
        return self._get(query_plan)

    def _get(self, query_plan: dict[DriverAgent, Set[PointNode]]):
        """Make query for selected points on each remote"""
        results, errors = {}, {}
        for (remote, point_set) in query_plan.items():
            q_return_values, q_return_errors = remote.get_multiple_points([p.identifier for p in point_set])
            for topic, val in q_return_values.items():
                node = self.equipment_tree.get_node(topic)
                if node:
                    node.last_value = val
            results.update(q_return_values)
            errors.update(q_return_errors)
        return results, errors

    @RPC.export
    def set(self, value: Any, topic: str | Sequence[str] | Set[str] = None, regex: str = None,
            confirm_values: bool = False, map_points: bool = False) -> tuple[dict, dict]:
        query_plan = self.build_query_plan(topic, regex)
        return self._set(value, query_plan, confirm_values, map_points)

    @RPC.export
    def semantic_set(self, value: Any, query: str, confirm_values: bool = False) -> tuple[dict, dict]:
        exact_matches = self.semantic_query(query)
        query_plan = self.build_query_plan(exact_matches)
        return self._set(value, query_plan, confirm_values)

    def _set(self, value: Any, query_plan: dict[DriverAgent, Set[PointNode]], confirm_values: bool, map_points=False
             ) -> tuple[dict, dict]:
        """Set selected points on each remote"""
        results, errors = {}, {}
        sender = self.vip.rpc.context.vip_message.peer
        for (remote, point_set) in query_plan.items():
            for point in point_set:
                try:
                    self.equipment_tree.raise_on_locks(point, sender)
                except:
                    pass # TODO: Handle this exception.
            # TODO: When map_points is True, all topics are sent to all remotes. This is probably wrong.
            point_value_tuples = list(value.items()) if map_points else [(p.identifier, value) for p in point_set]
            query_return_results, query_return_errors = remote.set_multiple_points(point_value_tuples)
            results.update(query_return_results)
            errors.update(query_return_errors)
            if confirm_values:
                # TODO: Should results contain the values read back from the device, or Booleans for success?
                results.update(remote.get_multiple_points([p.identifier for p in point_set])[0])
        return results, errors

    @RPC.export
    def revert(self, topic: str | Sequence[str] | Set[str] = None, regex: str = None) -> dict[str, str]:
              # confirm_values: bool = False) -> dict:
        query_plan = self.build_query_plan(topic, regex)
        return self._revert(query_plan)  # , confirm_values)

    @RPC.export
    def semantic_revert(self, query: str) -> dict[str, str]:  #, confirm_values: bool = False) -> dict:
        exact_matches = self.semantic_query(query)
        query_plan = self.build_query_plan(exact_matches)
        return self._revert(query_plan)  #, confirm_values)

    @staticmethod
    def _revert(query_plan) -> dict[str, str]:  #, confirm_values: bool) -> dict[str, str]:
        """
        Revert each point from query.
          If an exception is raised, return it in the error dict.
        """
        # TODO: If it is possible to check values, we may need to do that at the interface level.
        #  No functionality exists for this now.
        errors = {}
        for (remote, point_set) in query_plan.items():
            for point in point_set:
                try:
                    remote.revert_point(point.identifier)
                except Exception as e:
                    # TODO: revert_point may not raise. Does _set_point, typically?  If we make them raise,
                    #  we can return some errors, at least. It may not be possible to check success in all cases.
                    errors[point.identifier] = str(e)
        return errors

    @RPC.export
    def last(self, topic: str | Sequence[str] | Set[str] = None, regex: str = None,
             value: bool = True, updated: bool = True) -> dict:
        points = self.equipment_tree.find_points(topic, regex)
        return self._last(points, value, updated)

    @RPC.export
    def semantic_last(self, query: str, value: bool = True, updated: bool = True) -> dict:
        # 1. Get the list of topic strings
        exact_matches = self.semantic_query(query)

        # 2. Convert those topic strings into point node objects
        #    For example, if you have a helper method to do this:
        points = self.equipment_tree.find_points(exact_matches)

        # 3. Pass the point objects to _last
        return self._last(points, value, updated)

    @staticmethod
    def _last(points: Iterable[PointNode], value: bool, updated: bool):
        if value:
            if updated:
                return_dict = {p.topic: {
                    'value': p.last_value,
                    'updated': (p.last_updated.isoformat() if p.last_updated else None)} for p in points
                }
            else:
                return_dict = {p.topic: p.last_value for p in points}
        else:
            return_dict = {p.topic: (p.last_updated.isoformat() if p.last_updated else None) for p in points}
        return return_dict

    @RPC.export
    def call(self, method: str, topic: str | Sequence[str] | Set[str] = None, regex: str = None,
             *args, **kwargs) -> tuple[dict, dict]:
        # Find set of points to query and organize by remote:
        query_plan = self.build_query_plan(topic, regex)
        return self._call(query_plan, method, *args, **kwargs)

    @RPC.export
    def semantic_call(self, method: str, query: str, *args, **kwargs) -> tuple[dict, dict]:
        exact_matches = self.semantic_query(query)
        query_plan = self.build_query_plan(exact_matches)
        return self._call(query_plan, method, *args, **kwargs)

    @staticmethod
    def _call(query_plan: dict[DriverAgent, Set[PointNode]], method: str, *args, **kwargs):
        """Make query for selected points on each remote"""
        results, errors = {}, {}
        for (remote, point_set) in query_plan.items():
            q_return_values, q_return_errors = remote.call(method, [p.identifier for p in point_set], *args, **kwargs)
            results.update(q_return_values)
            errors.update(q_return_errors)
        return results, errors

    #-----------
    # UI Support
    #-----------
    @RPC.export
    def start(self, topic: str | Sequence[str] | Set[str] = None, regex: str = None) -> None:
        points = self.equipment_tree.find_points(topic, regex)
        self._start(points)

    @RPC.export
    def semantic_start(self, query: str) -> None:
        exact_matches = self.semantic_query(query)
        points = self.equipment_tree.find_points(exact_matches)
        self._start(points)

    def _start(self, points: Iterable[PointNode]) -> None:
        updates_required = []
        for p in points:
            if self.equipment_tree.is_active(p.identifier):
                continue
            else:
                p.active = True
                updates_required.append(p)
    # TODO: Add reschedule_all_on_update option and reschedule all poll_schedulers when true.
        if updates_required:
            self._update_polling_schedules(updates_required)

    @RPC.export
    def stop(self, topic: str | Sequence[str] | Set[str] = None, regex: str = None) -> None:
        points = self.equipment_tree.find_points(topic, regex)
        self._stop(points)

    @RPC.export
    def semantic_stop(self, query: str) -> None:
        topics = self.semantic_query(query)
        points = self.equipment_tree.find_points(topics)
        self._stop(points)

    def _stop(self, points: Iterable[PointNode]) -> None:
        for p in points:
            if not self.equipment_tree.is_active(p.identifier):
                continue
            else:
                p.active = False
                group = self.equipment_tree.get_group(p.identifier)
                self.poll_schedulers[group].remove_from_schedule(p, self.equipment_tree)
        # TODO: Add reschedule_all_on_update option and reschedule all poll_schedulers when true.

    @RPC.export
    def enable(self, topic: str | Sequence[str] | Set[str] = None, regex: str = None) -> None:
        nodes = self.equipment_tree.resolve_query(topic, regex)
        self._enable(nodes)

    @RPC.export
    def semantic_enable(self, query: str) -> None:
        topics = self.semantic_query(query)
        points = self.equipment_tree.resolve_query(topics)
        self._enable(points)

    def _enable(self, nodes: Iterable[DeviceNode | EquipmentNode | PointNode]):
        for node in nodes:
            node.config.active = True
            if not node.is_point:
                new_config = node.config.model_dump()
                if node.is_device:
                    self._add_fields_to_device_configuration_for_save(new_config, node)
                self.vip.config.set(node.identifier, new_config, trigger_callback=False)
            else:
                self.equipment_tree.update_stored_registry_config(node.identifier)

    @RPC.export
    def disable(self, topic: str | Sequence[str] | Set[str] = None, regex: str = None) -> None:
        nodes = self.equipment_tree.resolve_query(topic, regex)
        self._disable(nodes)

    @RPC.export
    def semantic_disable(self, query: str) -> None:
        topics = self.semantic_query(query)
        points = self.equipment_tree.resolve_query(topics)
        self._disable(points)

    def _disable(self, nodes: Iterable[DeviceNode | EquipmentNode | PointNode]) -> None:
        for node in nodes:
            node.config.active = False
            if not node.is_point:
                new_config = node.config.model_dump()
                if node.is_device:
                    self._add_fields_to_device_configuration_for_save(new_config, node)
                self.vip.config.set(node.identifier, new_config, trigger_callback=False)
            else:
                self.equipment_tree.update_stored_registry_config(node.identifier)

    def _add_fields_to_device_configuration_for_save(self, new_config, node):
        registry_name = node.registry_name
        if not registry_name or not (registry_name := self.equipment_tree.set_registry_name(node.identifier)):
            raise Exception(f'Unable to set configuration for device node {node.identifier}.'
                            f' Registry name is unknown and cannot be determined.')
        # TODO: This assumes that the registry was originally provided as a separate file.
        #  We should detect this and modify the file or dict that was originally configured.
        new_config['registry_config'] = f'config://{registry_name}'
        new_config['remote_config'] = self.equipment_tree.get_remote(node.identifier).config.model_dump()

    @RPC.export
    def status(self, topic: str | Sequence[str] | Set[str] = None, regex: str = None) -> dict:
        nodes = self.equipment_tree.find_points(topic, regex)
        return self._status(nodes)

    @RPC.export
    def semantic_status(self, query: str) -> dict:
        topics = self.semantic_query(query)
        points = self.equipment_tree.find_points(topics)
        return self._status(points)

    def _status(self, points: Iterable[PointNode]) -> dict:
        raise NotImplementedError('status is not yet implemented.')
        # # TODO: Implement _status()
        # return {'error': 'Status reporting is not yet implemented'}

    @RPC.export
    def add_node(self, node_topic: str, config: dict, update_schedule: bool = True) -> bool:
        # TODO: Need logic to determine if this is a point. Configure_new_equipment should not be used if it is.
        return self._configure_new_equipment(node_topic, 'NEW', contents=config, schedule_now=update_schedule)

    @RPC.export
    def remove_node(self, node_topic: str, leave_disconnected: bool = False) -> bool:
        return self._remove_equipment(node_topic, None, None, leave_disconnected)

    @RPC.export
    def add_interface(self, interface_name: str, force: bool = False, pre_release: bool = False) -> bool:
        # TODO: vdrv might want to get the absolute path before calling this, if it is a path.
        interface_path = Path(interface_name)
        if interface_path.is_dir():
            with open(interface_path / 'pyproject.toml', 'rb') as f:
                ppt = tomllib.load(f)
            package_name, install_path = ppt['tool']['poetry']['name'], str(interface_path)
        elif interface_path.is_file() and interface_path.suffix == ".whl":
            package_name, install_path = str(interface_path.stem).split('-')[0], str(interface_path)
        else:  # Use package name to get it from PyPI.
            package_name = install_path = self._interface_package_from_short_name(interface_name)
        try:
            distribution('volttron-core')
            from functools import partial
            # TODO: This method should probably be moved at least mostly into vctl,
            #  and it should call the control service directly.
            #  Otherwise, this may try to package a directory which is on the client here on the server.
            class Conn:
                call = partial(self.vip.rpc.call, 'platform.control')
            arguments = Namespace(connection=Conn(), install_path=install_path,
                                  force=force, pre_release=pre_release)
            try:
                with redirect_stdout(st_out := StringIO()):
                    install_lib_vctl(arguments)
                success = True if st_out.getvalue().startswith('Installed') else False
            except ValueError as e:
                _log.warning(f'Failed to install interface "{interface_name}": {e}')
                success = False
        except PackageNotFoundError:  # Monolithic does not have install-lib. Handle this manually.
            try:
                subprocess.run(['pip', 'install', '--no-deps', str(install_path)],
                               check=True, capture_output=True)
                exclude_packages = ['python', 'volttron-core', 'volttron-lib-base-driver']
                if deps := [d for d in requires(package_name) if d.split(' ')[0] not in exclude_packages]:
                    subprocess.run((['pip', 'install', *deps]), check=True, capture_output=True)
                success = True
            except subprocess.CalledProcessError as e:
                _log.warning(f'Failed to install interface "{interface_name}": {e.stderr}')
                success = False
        if success:
            _log.info(f'Successfully installed {interface_name} driver interface.')
        return success

    @RPC.export
    def list_interfaces(self) -> list[str]:
        """Return list of all installed driver interfaces."""
        try:
            from volttron.driver import interfaces
            return [i.name for i in iter_modules(interfaces.__path__)]
        except ImportError:
            return []

    @RPC.export
    def remove_interface(self, interface_name: str) -> bool:
        interface_package = self._interface_package_from_short_name(interface_name)
        try:
            distribution('volttron-core')
            subprocess.run(['poetry', '--directory', 'remove', interface_package], capture_output=True)
        except PackageNotFoundError:
            subprocess.run([sys.executable, '-m', 'pip', 'uninstall', interface_package, '--no-input'],
                           capture_output=True)
        sp_result = subprocess.run([sys.executable, '-m', 'pip', 'show', interface_package], capture_output=True)
        return True if sp_result.returncode else False

    @RPC.export
    def list_topics(self, topic: str, regex: str = None,
                    active: bool = False, enabled: bool = False) -> list[str]:
        # TODO: Semantic version?
        topic = topic.strip('/') if topic and topic.startswith(self.equipment_tree.root) else self.equipment_tree.root
        parent = topic if self.equipment_tree.get_node(topic) else topic.rsplit('/', 1)[0]
        children = self.equipment_tree.children(parent)
        regex = re.compile(regex) if regex else None
        if regex:
            children = [c for c in children if regex.search(c)]
        if active:
            children = [c for c in children if self.equipment_tree.is_active(c.identifier)]
        if enabled:
            children = [c for c in children if c.enabled]
        return [c.identifier for c in children]

    @RPC.export
    def get_poll_schedule(self, full_topics=False):
        return {group: scheduler.get_schedule(full_topics) for group, scheduler in self.poll_schedulers.items()}

    @RPC.export
    def export_equipment_tree(self):
        return self.equipment_tree.to_json(with_data=True)

    #-------------
    # Reservations
    #-------------
    # TODO: Improve the Reservations and Overrides API:
    # @RPC.export
    # def new_reservation(self, task_id: str, priority: str, requests: list) -> dict|None:
    #     """
    #     Reserve one or more blocks on time on one or more device.
    #
    #     :param task_id: An identifier for this reservation.
    #     :param priority: Priority of the task. Must be either "HIGH", "LOW",
    #     or "LOW_PREEMPT"
    #     :param requests: A list of time slot requests in the format
    #     described in `Device Schedule`_.
    #     """
    #     rpc_peer = self.vip.rpc.context.vip_message.peer
    #     return self.reservation_manager.new_task(rpc_peer, task_id, priority, requests)  #, publish_result=False)
    #
    # @RPC.export
    # def cancel_reservation(self, task_id: str) -> dict|None:
    #     """
    #     Requests the cancellation of the specified task id.
    #     :param task_id: Task name.
    #     """
    #     rpc_peer = self.vip.rpc.context.vip_message.peer
    #     return self.reservation_manager.cancel_task(rpc_peer, task_id)  # , publish_result=False)

    #----------
    # Overrides
    #----------
    @RPC.export
    def set_override_on(self, pattern: str, duration: float = 0.0,
                        failsafe_revert: bool = True, staggered_revert: bool = False):
        """RPC method

        Turn on override condition on all the devices matching the pattern.
        :param pattern: Override pattern to be applied. For example,
            If pattern is campus/building1/* - Override condition is applied for all the devices under
            campus/building1/.
            If pattern is campus/building1/ahu1 - Override condition is applied for only campus/building1/ahu1
            The pattern matching is based on bash style filename matching semantics.
        :type pattern: str
        :param duration: Time duration for the override in seconds. If duration <= 0.0, it implies as indefinite
        duration.
        :type duration: float
        :param failsafe_revert: Flag to indicate if all the devices falling under the override condition has to be set
         to its default state/value immediately.
        :type failsafe_revert: boolean
        :param staggered_revert: If this flag is set, reverting of devices will be staggered.
        :type staggered_revert: boolean
        """
        self.override_manager.set_on(pattern, duration, failsafe_revert, staggered_revert)

    @RPC.export
    def set_override_off(self, pattern: str):
        """RPC method

        Turn off override condition on all the devices matching the pattern. The pattern matching is based on bash style
        filename matching semantics.
        :param pattern: Pattern on which override condition has to be removed.
        :type pattern: str
        """
        return self.override_manager.set_off(pattern)

    # Get a list of all the devices with override condition.
    @RPC.export
    def get_override_devices(self):
        """RPC method

        Get a list of all the devices with override condition.
        """
        return list(self.override_manager.devices)

    @RPC.export
    def clear_overrides(self):
        """RPC method

        Clear all overrides.
        """
        self.override_manager.clear()

    @RPC.export
    def get_override_patterns(self):
        """RPC method

        Get a list of all the override patterns.
        """
        return list(self.override_manager.patterns)

    #-------------------
    # Legacy RPC Methods
    #-------------------
    @RPC.export
    def get_point(self, path: str = None, point_name: str = None, **kwargs) -> Any:
        """
        RPC method

        Gets up-to-date value of a specific point on a device.
        Does not require the device be scheduled.

        :param path: The topic of the point to grab in the
                      format <device topic>/<point name>

                      Only the <device topic> if point is specified.
        :param point_name: Point on the device. Assumes topic includes point name if omitted.
        :param kwargs: Any driver specific parameters
        :type path: str
        :returns: point value
        :rtype: any base python type"""

        # Support for old-actuator-style keyword arguments.
        path = path if path else kwargs.get('topic', None)
        point_name = point_name if point_name else kwargs.get('point', None)
        if path is None:
            # DEPRECATED: Only allows topic to be None to permit use of old-actuator-style keyword argument "topic".
            raise TypeError('Argument "path" is required.')

        point_name = self._equipment_id(path, point_name)
        node = self.equipment_tree.get_node(point_name)
        if not node:
            raise ValueError(f'No equipment found for topic: {point_name}')
        remote = self.equipment_tree.get_remote(node.identifier)
        if not remote:
            raise ValueError(f'No remote found for topic: {point_name}')
        return remote.get_point(point_name, **kwargs)

    @RPC.export
    def set_point(self, path: str, point_name: str | None, value: Any, *args, **kwargs) -> Any:
        """RPC method

        Sets the value of a specific point on a device.
        Requires the device be scheduled by the calling agent.

        :param path: The topic of the point to set in the
                      format <device topic>/<point name>
                      Only the <device topic> if point is specified.
        :param value: Value to set point to.
        :param point_name: Point on the device.
        :param kwargs: Any driver specific parameters
        :type path: str
        :type value: any basic python type
        :type point_name: str
        :returns: value point was actually set to. Usually invalid values
                cause an error but some drivers (MODBUS) will return a
                different
                value with what the value was actually set to.
        :rtype: any base python type

        .. warning:: Calling will raise a ReservationLockError if another agent has already scheduled
        this device for the present time."""

        sender = self.vip.rpc.context.vip_message.peer

        # Support for old-actuator-style arguments.
        topic = kwargs.get('topic')
        if topic:
            path = topic
        elif path == sender or len(args) > 0:
            # Function was likely called with actuator-style positional arguments. Reassign variables to match.
            _log.info('Deprecated actuator-style positional arguments detected in set_point().'
                      ' Please consider converting code to use set() method.')
            path, point_name = (point_name, args[0]) if len(args) >= 1 else point_name, None
        point_name = point_name if point_name else kwargs.get('point', None)

        point_name = self._equipment_id(path, point_name)
        return self._set_point(point_name, value, sender, **kwargs)

    def _set_point(self, topic, value, sender, **kwargs):
        node: EquipmentNode = self.equipment_tree.get_node(topic)
        if not node:
            raise ValueError(f'No equipment found for topic: {topic}')
        self.equipment_tree.raise_on_locks(node, sender)
        remote = self.equipment_tree.get_remote(node.identifier)
        if not remote:
            raise ValueError(f'No remote found for topic: {topic}')
        result = remote.set_point(topic, value, **kwargs)
        headers = self._get_headers(sender)
        self._push_result_topic_pair(WRITE_ATTEMPT_PREFIX, topic, headers, value)
        self._push_result_topic_pair(VALUE_RESPONSE_PREFIX, topic, headers, result)
        return result

    @RPC.export
    def scrape_all(self, topic: str) -> dict:
        """RPC method

        Get all points from a device.

        :param topic: Device topic
        :returns: Dictionary of points to values
        """
        _log.info('Call to deprecated RPC method "scrape_all". This method has been superseded by the "get" method'
                  ' and will be removed in a future version. Please update to the newer method.')
        path = self._equipment_id(topic, None)
        return self.get(topic=path)[0]

    @RPC.export
    def get_multiple_points(self, path: str | Sequence[str | Sequence] = None, point_names = None,
                            **kwargs) -> tuple[dict, dict]:
        """RPC method

        Get multiple points on multiple devices. Makes a single
        RPC call to the platform driver per device.

        :param path: A topic (with or without point names), a list of full topics (with point names),
         or a list of [device, point] pairs.
        :param point_names: A Sequence of point names associated with the given path.
        :param kwargs: Any driver specific parameters

        :returns: Dictionary of points to values and dictionary of points to errors

        .. warning:: This method does not require that all points be returned
                     successfully. Check that the error dictionary is empty.
        """

        # Support for actuator-style keyword arguments.
        topics = path if path else kwargs.get('topics', None)
        if topics is None:
            # path is allowed to be None to permit use of old-actuator-style keyword argument "topics".
            raise TypeError('Argument "path" is required.')

        errors = {}
        devices = set()
        if isinstance(topics, str):
            if not point_names:
                devices.add(topics)
            else:
                for point in point_names:
                    devices.add(self._equipment_id(topics, point))
        elif isinstance(topics, Sequence):
            for topic in topics:
                if isinstance(topic, str):
                    devices.add(self._equipment_id(topic))
                elif isinstance(topic, Sequence) and len(topic) == 2:
                    devices.add(self._equipment_id(*topic))
                else:
                    e = ValueError("Invalid topic: {}".format(topic))
                    errors[repr(topic)] = repr(e)

        results, query_errors = self.get(devices)
        errors.update(query_errors)
        return results, errors

    @RPC.export
    def set_multiple_points(self, path: str, point_names_values: list[tuple[str, Any]], **kwargs) -> dict:
        """RPC method

        Set values on multiple set points at once. If global override is condition is set,raise OverrideError exception.
        :param path: device path
        :type path: str
        :param point_names_values: list of points and corresponding values
        :type point_names_values: list of tuples
        :param kwargs: additional arguments for the device
        :type kwargs: arguments pointer
        """
        errors = {}
        topic_value_map = {}
        sender = self.vip.rpc.context.vip_message.peer
        # Support for old-actuator-style positional arguments so long as sender matches rpc peer.
        topics_values = kwargs.get('topics_values')
        if path == sender or topics_values is not None:  # Method was called with old-actuator-style arguments.
            topics_values = topics_values if topics_values else point_names_values
            for topic, value in topics_values:
                if isinstance(topic, str):
                    topic_value_map[self._equipment_id(topic, None)] = value
                elif isinstance(topic, Sequence) and len(topic) == 1:
                    topic_value_map[self._equipment_id(*topic)] = value
                else:
                    e = ValueError("Invalid topic: {}".format(topic))
                    errors[str(topic)] = repr(e)
        else:  # Assume method was called with old-driver-style arguments.
            for point, value in point_names_values:
                topic_value_map[self._equipment_id(path, point)] = value

        _, ret_errors = self.set(topic_value_map, map_points=True, **kwargs)
        errors.update(ret_errors)
        return errors

    def heart_beat(self):
        """
        Sends heartbeat to all devices
        """
        # TODO: This should send to all remotes, collect tasks, and then log tasks after the fact.
        # TODO: Move this into the PollScheduler with configurable (per device) set of points and intervals (per-point).
        _log.debug("sending heartbeat")
        for remote in self.equipment_tree.remotes.values():
            try:
                remote.heart_beat()
            except (Exception, gevent.Timeout) as e:
                _log.warning(f'Failed to set heart_beat point on remote: {remote.unique_id} -- {e}.')

    @RPC.export
    def revert_point(self, path: str, point_name: str, **kwargs):
        """RPC method

        Revert the set point to default state/value.
        If global override is condition is set, raise OverrideError exception.
        If topic has been reserved by another user
        or if it is not reserved but reservations are required,
         raise ReservationLockError exception.
        :param path: device path
        :type path: str
        :param point_name: set point to revert
        :type point_name: str
        :param kwargs: additional arguments for the device
        :type kwargs: arguments pointer
        """
        sender = self.vip.rpc.context.vip_message.peer

        # Support for old-actuator-style arguments.
        topic = kwargs.get('topic')
        if topic:
            path, point_name = topic, None
        elif path == sender:
            # Function was likely called with actuator-style positional arguments. Reassign variables to match.
            _log.info('Deprecated actuator-style positional arguments detected in revert_point().'
                      ' Please consider converting code to use revert() method.')
            path, point_name = point_name, None

        equip_id = self._equipment_id(path, point_name)
        node = self.equipment_tree.get_node(equip_id)
        if not node:
            raise ValueError(f'No equipment found for topic: {equip_id}')
        self.equipment_tree.raise_on_locks(node, sender)
        remote = self.equipment_tree.get_remote(node.identifier)
        remote.revert_point(equip_id, **kwargs)

        headers = self._get_headers(sender)
        self._push_result_topic_pair(REVERT_POINT_RESPONSE_PREFIX, equip_id, headers, None)

    @RPC.export
    def revert_device(self, path: str, *args, **kwargs):
        """RPC method

        Revert all the set point values of the device to default state/values. If global override is condition is set,
        raise OverrideError exception.
        :param path: device path
        :type path: str
        :param kwargs: additional arguments for the device
        :type kwargs: arguments pointer
        """
        sender = self.vip.rpc.context.vip_message.peer

        # Support for old-actuator-style arguments.
        topic = kwargs.get('topic')
        if topic:
            path = topic
        elif path == sender and len(args) > 0:
            # Function was likely called with actuator-style positional arguments. Reassign variables to match.
            _log.info('Deprecated actuator-style positional arguments detected in revert_device().'
                      ' Please consider converting code to use revert() method.')
            path = args[0]

        self.revert(self._equipment_id(path, None))

        headers = self._get_headers(sender)
        self._push_result_topic_pair(REVERT_DEVICE_RESPONSE_PREFIX, path, headers, None)

    @RPC.export
    def request_new_schedule(self, _, task_id: str, priority: str,
                             requests: list[list[str]] | list[str], **__) -> dict:
        """
        RPC method

        Requests one or more blocks on time on one or more device.

        :param _: formerly requester_id -- now ignored, VIP Identity used internally
        :param task_id: Task name.
        :param priority: Priority of the task. Must be either "HIGH", "LOW",
        or "LOW_PREEMPT"
        :param requests: A list of time slot requests in the format
        described in `Device Schedule`_.

        :type priority: str
        :returns: Request result
        :rtype: dict

        :return Values:

            The return values are described in `New Task Response`_.
        """
        # _log.info('Call to deprecated RPC method "request_new_schedule". '
        #            'This method provides compatability with the actuator API, but has been superseded '
        #            'by "new_reservation". Please update to the newer method.')
        rpc_peer = self.vip.rpc.context.vip_message.peer
        return self.reservation_manager.new_task(rpc_peer, task_id, priority, requests)

    @RPC.export
    def request_cancel_schedule(self, _, task_id: str, **__) -> dict:
        """RPC method

        Requests the cancellation of the specified task id.

        :param _: formerly requester_id -- now ignored, VIP Identity used internally
        :param task_id: Task name.

        :returns: Request result
        :rtype: dict

        :return Values:

        The return values are described in `Cancel Task Response`_.

        """
        # _log.info('Call to deprecated RPC method "request_cancel_schedule". '
        #            'This method provides compatability with the actuator API, but has been superseded '
        #            'by "cancel_reservation". Please update to the newer method.')
        rpc_peer = self.vip.rpc.context.vip_message.peer
        return self.reservation_manager.cancel_task(rpc_peer, task_id)

    ##################
    # PubSub Interface
    ##################

    def handle_get(self, _, sender: str, __, topic: str, ___, ____):
        """
        Requests up-to-date value of a point.

        To request a value publish a message to the following topic:

        ``devices/actuators/get/<device path>/<actuation point>``

        with the fallowing header:

        .. code-block:: python

            {
                'requesterID': <Ignored, VIP Identity used internally>
            }

        The ActuatorAgent will reply on the **value** topic
        for the actuator:

        ``devices/actuators/value/<full device path>/<actuation point>``

        with the message set to the value the point.

        """
        point = topic.replace(GET_TOPIC + '/', '', 1)
        headers = self._get_headers(sender)
        try:
            value = self.get_point(point)
            self._push_result_topic_pair(VALUE_RESPONSE_PREFIX, point, headers, value)
        except Exception as ex:
            self._handle_error(ex, point, headers)

    def handle_set(self, _, sender: str, __, topic: str, ___, message: Any):
        """
        Set the value of a point.

        To set a value publish a message to the following topic:

        ``devices/actuators/set/<device path>/<actuation point>``

        with the fallowing header:

        .. code-block:: python

            {
                'requesterID': <Ignored, VIP Identity used internally>
            }

        The ActuatorAgent will reply on the **value** topic
        for the actuator:

        ``devices/actuators/value/<full device path>/<actuation point>``

        with the message set to the value the point.

        Errors will be published on

        ``devices/actuators/error/<full device path>/<actuation point>``

        with the same header as the request.

        """
        point = topic.replace(SET_TOPIC + '/', '', 1)
        headers = self._get_headers(sender)
        if not message:
            error = {'type': 'ValueError', 'value': 'missing argument'}
            _log.debug('ValueError: ' + str(error))
            self._push_result_topic_pair(ERROR_RESPONSE_PREFIX, point, headers, error)
            return

        try:
            equip_id = self._equipment_id(point)
            self._set_point(equip_id, message, sender, **{})
        except Exception as ex:
            self._handle_error(ex, point, headers)

    def handle_revert_point(self, _, sender: str, __, topic: str, ___, ____):
        """
        Revert the value of a point.

        To revert a value publish a message to the following topic:

        ``actuators/revert/point/<device path>/<actuation point>``

        with the following header:

        .. code-block:: python

            {
                'requesterID': <Ignored, VIP Identity used internally>
            }

        The ActuatorAgent will reply on

        ``devices/actuators/reverted/point/<full device path>/<actuation point>``

        to indicate that a point was reverted.

        Errors will be published on

        ``devices/actuators/error/<full device path>/<actuation point>``

        with the same header as the request.
        """
        topic = self._equipment_id(topic.replace(REVERT_POINT_TOPIC + '/', '', 1), None)
        headers = self._get_headers(sender)
        try:
            node = self.equipment_tree.get_node(topic)
            if not node:
                raise ValueError(f"No point node found for topic: {topic}")

            self.equipment_tree.raise_on_locks(node, sender)
            remote = self.equipment_tree.get_remote(node.identifier)
            if not remote:
                raise ValueError(f"No remote found for point: {topic}")

            remote.revert_point(topic)

            self._push_result_topic_pair(REVERT_POINT_RESPONSE_PREFIX, topic, headers, None)
        except Exception as ex:
            self._handle_error(ex, topic, headers)

    def handle_revert_device(self, _, sender: str, __, topic: str, ___, ____):
        """
        Revert all the writable values on a device.

        To revert a device publish a message to the following topic:

        ``devices/actuators/revert/device/<device path>``

        with the following header:

        .. code-block:: python

            {
                'requesterID': <Ignored, VIP Identity used internally>
            }

        The ActuatorAgent will reply on

        ``devices/actuators/reverted/device/<full device path>``

        to indicate that a device was reverted.

        Errors will be published on

        ``devices/actuators/error/<full device path>/<actuation point>``

        with the same header as the request.
        """
        topic = topic.replace(REVERT_DEVICE_TOPIC + '/', '', 1)
        topic = self._equipment_id(topic, None)
        headers = self._get_headers(sender)
        try:
            device_node = self.equipment_tree.get_node(topic)
            if not device_node:
                raise ValueError(f"No device node found for topic: {topic}")

            self.equipment_tree.raise_on_locks(device_node, sender)
            self.revert(device_node.identifier)

            self._push_result_topic_pair(REVERT_DEVICE_RESPONSE_PREFIX, topic, headers, None)

        except Exception as ex:
            self._handle_error(ex, topic, headers)

    def handle_reservation_request(self, _, sender: str, __, topic: str, headers: dict,
                                   message: list[list[str]] | list[str]):
        """
        Schedule request pub/sub handler

        An agent can request a task schedule by publishing to the
        ``devices/actuators/schedule/request`` topic with the following header:

        .. code-block:: python

            {
                'type': 'NEW_SCHEDULE',
                'requesterID': <Ignored, VIP Identity used internally>,
                'taskID': <unique task ID>, #The desired task ID for this
                task. It must be unique among all scheduled tasks.
                'priority': <task priority>, #The desired task priority,
                must be 'HIGH', 'LOW', or 'LOW_PREEMPT'
            }

        The message must describe the blocks of time using the format
        described in `Device Schedule`_.

        A task may be canceled by publishing to the
        ``devices/actuators/schedule/request`` topic with the following header:

        .. code-block:: python

            {
                'type': 'CANCEL_SCHEDULE',
                'requesterID': <Ignored, VIP Identity used internally>,
                'taskID': <unique task ID>, #The task ID for the canceled Task.
            }

        requesterID
            The name of the requesting agent. Automatically replaced with VIP id.
        taskID
            The desired task ID for this task. It must be unique among all
            scheduled tasks.
        priority
            The desired task priority, must be 'HIGH', 'LOW', or 'LOW_PREEMPT'

        No message is required to cancel a schedule.

        """
        request_type = headers.get('type')
        _log.debug(f'handle_schedule_request: {topic}, {headers}, {message}')

        task_id = headers.get('taskID')
        priority = headers.get('priority')

        now = get_aware_utc_now()
        if request_type == RESERVATION_ACTION_NEW or request_type == LEGACY_RESERVATION_ACTION_NEW:
            try:
                requests = message[0] if len(message) == 1 else message
                headers = self._get_headers(sender, now, task_id, RESERVATION_ACTION_NEW)
                result = self.reservation_manager.new_task(sender, task_id, priority, requests, now)
            except Exception as ex:
                return self._handle_unknown_reservation_error(ex, headers, message)
            # Dealing with success and other first world problems.
            if result.success:
                for preempted_task in result.data:
                    preempt_headers = self._get_headers(preempted_task[0], task_id=preempted_task[1],
                                                        action_type=RESERVATION_ACTION_CANCEL)
                    self.vip.pubsub.publish('pubsub',
                                            topic=RESERVATION_RESULT_TOPIC,
                                            headers=preempt_headers,
                                            message={
                                                'result': RESERVATION_CANCEL_PREEMPTED,
                                                'info': '',
                                                'data': {
                                                    'agentID': sender,
                                                    'taskID': task_id
                                                }
                                            })
            results = {'result': (RESERVATION_RESPONSE_SUCCESS if result.success else RESERVATION_RESPONSE_FAILURE),
                       'data': (result.data if not result.success else {}),
                       'info': result.info_string}
            self.vip.pubsub.publish('pubsub', topic=RESERVATION_RESULT_TOPIC, headers=headers, message=results)


        elif request_type == RESERVATION_ACTION_CANCEL or request_type == LEGACY_RESERVATION_ACTION_CANCEL:
            try:
                result = self.reservation_manager.cancel_reservation(sender, task_id)
                message = {
                    'result': (RESERVATION_RESPONSE_SUCCESS if result.success else RESERVATION_RESPONSE_FAILURE),
                    'info': result.info_string,
                    'data': {}
                }
                topic = RESERVATION_RESULT_TOPIC
                headers = self._get_headers(sender, now, task_id, RESERVATION_ACTION_CANCEL)
                self.vip.pubsub.publish('pubsub', topic, headers=headers, message=message)

            except Exception as ex:
                return self._handle_unknown_reservation_error(ex, headers, message)
        else:
            _log.debug('handle-schedule_request, invalid request type')
            self.vip.pubsub.publish('pubsub', RESERVATION_RESULT_TOPIC, headers, {
                'result': RESERVATION_RESPONSE_FAILURE,
                'info': 'INVALID_REQUEST_TYPE',
                'data': {}
            })

    ################
    # Helper Methods
    ################

    def _equipment_id(self, path: str, point: str = None) -> str:
        path = path.strip('/')
        if point is not None:
            path = '/'.join([path, point])
        # If path already starts with "devices/", skip prefixing
        if not path.startswith(self.equipment_tree.root + '/'):
            path = '/'.join([self.equipment_tree.root, path])
        return path

    @staticmethod
    def _get_headers(requester: str, time: datetime = None, task_id: str = None, action_type: str = None):
        # TODO: This method should have a better name that reflects its actuator-specific usage.
        headers = {'time': format_timestamp(time) if time else format_timestamp(get_aware_utc_now())}
        if requester is not None:
            headers['requesterID'] = requester
        if task_id is not None:
            headers['taskID'] = task_id
        if action_type is not None:
            headers['type'] = action_type
        return headers

    def _handle_error(self, ex: BaseException, point: str, headers: dict):
        if isinstance(ex, RemoteError):
            try:
                exc_type = ex.exc_info['exc_type']
                exc_args = ex.exc_info['exc_args']
            except KeyError:
                exc_type = "RemoteError"
                exc_args = ex.message
            error = {'type': exc_type, 'value': str(exc_args)}
        else:
            error = {'type': ex.__class__.__name__, 'value': str(ex)}
        self._push_result_topic_pair(ERROR_RESPONSE_PREFIX, point, headers, error)
        _log.warning('Error handling subscription: ' + str(error))

    def _handle_unknown_reservation_error(self, ex: BaseException, headers: dict, message: list[list[str]] | list[str]):
        _log.warning(f'bad request: {headers}, {message}, {str(ex)}')
        results = {
            'result': "FAILURE",
            'data': {},
            'info': 'MALFORMED_REQUEST: ' + ex.__class__.__name__ + ': ' + str(ex)
        }
        self.vip.pubsub.publish('pubsub', RESERVATION_RESULT_TOPIC, headers=headers, message=results)
        return results

    @staticmethod
    def _interface_package_from_short_name(interface_name):
        if interface_name.startswith('volttron-lib-') and interface_name.endswith('-driver'):
            return interface_name
        else:
            return f'volttron-lib-{interface_name}-driver'

    def _push_result_topic_pair(self, prefix: str, point: str, headers: dict, value: Any):
        topic = normtopic('/'.join([prefix, point]))
        self.vip.pubsub.publish('pubsub', topic, headers, message=value)

    def _split_topic(self, topic: str, point: str = None) -> tuple[str, str]:
        """Convert actuator-style optional point names to (path, point) pair."""
        topic = topic.strip('/')
        if not topic.startswith(self.equipment_tree.root):
            topic = '/'.join([self.equipment_tree.root, topic])
        path, point_name = (topic, point) if point is not None else topic.rsplit('/', 1)
        return path, point_name

    @RPC.export
    def forward_bacnet_cov_value(self, remote_id, topic, point_values):
        """
        Called by the BACnet Proxy to pass the COV value to the driver agent
        for publishing
        :param remote_id: The unique addressable identifier of the remote.
        :param topic: name of the point in the COV notification
        :param point_values: dictionary of updated values sent by the device
        """
        self.equipment_tree.remotes[remote_id].publish_cov_value(topic, point_values)


def main():
    """Main method called to start the agent."""
    vip_main(PlatformDriverAgent, identity=PLATFORM_DRIVER, version=__version__)


if __name__ == '__main__':
    # Entry point for script
    try:
        sys.exit(main())
    except KeyboardInterrupt:
        pass



# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

import logging
import sys

from datetime import datetime
from importlib.metadata import distribution, PackageNotFoundError

try:
    distribution('volttron-core')
    from volttron.utils.math_utils import mean, stdev
except PackageNotFoundError:
    from volttron.platform.agent.math_utils import mean, stdev

_log = logging.getLogger(__name__)


# TODO: Does this design still work for the new driver?
class ScalabilityTester:
    def __init__(self, required_iterations):
        self.required_iterations = required_iterations

        self.waiting_to_finish = set()
        self.test_iterations = 0
        self.test_results = []
        self.current_test_start = None

    def poll_starting(self, topic):
        if not self.waiting_to_finish:
            # Start a new measurement
            self.current_test_start = datetime.now()
            # TODO: What should be in this set for the new polling architecture?
            self.waiting_to_finish = set(self.instances.keys())

        if topic not in self.waiting_to_finish:
            _log.warning(
                f"{topic} started twice before test finished, increase the length of scrape interval and rerun test"
            )

    def poll_ending(self, topic):
        try:
            self.waiting_to_finish.remove(topic)
        except KeyError:
            _log.warning(
                f"{topic} published twice before test finished, increase the length of scrape interval and rerun test"
            )

        if not self.waiting_to_finish:
            end = datetime.now()
            delta = end - self.current_test_start
            delta = delta.total_seconds()
            self.test_results.append(delta)

            self.test_iterations += 1

            _log.info("publish {} took {} seconds".format(self.test_iterations, delta))

            if self.test_iterations >= self.required_iterations:
                # Test is now over. Button it up and shutdown.
                mean_t = mean(self.test_results)
                stdev_t = stdev(self.test_results)
                _log.info("Mean total publish time: " + str(mean_t))
                _log.info("Std dev publish time: " + str(stdev_t))
                sys.exit(0)



#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK
from importlib.metadata import distribution, PackageNotFoundError

import argcomplete
import argparse
import logging
import os
import sys

from pprint import pprint

try:
    distribution('volttron-core')
    from volttron.client.commands.connection import ControlConnection
    from volttron.client.logs import AgentFormatter
    from volttron.utils import parse_json_config, ClientContext as cc
    connection = ControlConnection(cc.get_address(), peer='platform.driver')
except PackageNotFoundError:
    from volttron.platform import get_address
    from volttron.platform.agent.utils import parse_json_config, AgentFormatter
    from volttron.platform.control.control_connection import ControlConnection
    from volttron.platform.vip.agent.utils import build_agent
    connection = ControlConnection(get_address(), peer='platform.driver')

_log = logging.getLogger(os.path.basename(sys.argv[0]) if __name__ == "__main__" else __name__)


# TODO: How to supress space after topic-part is completed?
def topic_completer(prefix, **kwargs):
    topic = prefix if prefix else None
    return connection.call('list_topics', topic=topic)

def driver_get_func(args):
    result, errors = connection.call('get', topic=args.topic, regex=args.regex)
    return (result, errors) if not errors and not result else result if not errors else result

def driver_set_func(args):
    result, errors = connection.call('set',  value=args.value, topic=args.topic, regex=args.regex,
                                     confirm_values=args.confirm_values, map_points=args.map_points)
    #TODO: This never seems to return errors. Why? Fix this and similar.
    return (result, errors) # if not errors and not result else result if not errors else errors

def driver_last_func(args):
    return connection.call('last', topic=args.topic, regex=args.regex,
                           value=args.value, updated=args.updated)

def driver_revert_func(args):
    return connection.call('revert', topic=args.topic, regex=args.regex)

def driver_start_func(args):
    return connection.call('start', topic=args.topic, regex=args.regex)

def driver_stop_func(args):
    return connection.call('stop', topic=args.topic, regex=args.regex)

def driver_enable_func(args):
    return connection.call('enable', topic=args.topic, regex=args.regex)

def driver_disable_func(args):
    return connection.call('disable', topic=args.topic, regex=args.regex)

def driver_status_func(args):
    return connection.call('status', topic=args.topic, regex=args.regex)

def driver_node_add_func(args):
    if not args.config and args.config_path:
        with open(args.config_path) as f:
            config = parse_json_config(f)
    else:
        config = args.config
    return connection.call('add_node', node_topic=args.topic, config=config, update_schedule=args.update_schedule)

def driver_node_remove_func(args):
    return connection.call('remove_node', topic=args.topic)

def driver_interface_install_func(args):
    return connection.call('add_interface', interface_name=args.name)

def driver_interface_list_func(args):
    return connection.call('list_interfaces')

def driver_interface_remove_func(args):
    return connection.call('remove_interface', interface_name=args.name)

def driver_list_topics_func(args):
    return connection.call('list_topics', topic=args.topic, regex=args.regex)

def main():
    if not connection.peer in connection.server.vip.peerlist().get():
        sys.stderr.write('Platform Driver Agent must be running to use driver subcommands.\n')
        sys.exit(1)
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers()
    parser_driver = subparsers.add_parser('driver', help='Driver Commands')
    driver_subparsers = parser_driver.add_subparsers()

    # TODO: How to allow empty topic, but still have it be positional?
    # TODO: How to suppress optional arguments from suggestions?
    parser_driver_get = driver_subparsers.add_parser('get', help='Query data from equipment.')
    parser_driver_get.add_argument('topic', help='Topic or pattern to select equipment.').completer = topic_completer
    parser_driver_get.add_argument('-r', '--regex', help='Regex pattern to modify equipment selection.')
    parser_driver_get.add_argument('-t', '--tag', help='Tagging expression to select equipment')
    parser_driver_get.set_defaults(func=driver_get_func)

    parser_driver_set = driver_subparsers.add_parser('set', help='Set data on equipment.')
    parser_driver_set.add_argument('topic', help='Topic or pattern to select equipment.').completer = topic_completer
    parser_driver_set.add_argument('value', help='Value to which to set the selected points.')
    parser_driver_set.add_argument('-r', '--regex', help='Regex pattern to modify equipment selection.')
    parser_driver_set.add_argument('-t', '--tag', help='Tagging expression to select equipment.')
    parser_driver_set.add_argument('--confirm-values', help='Query points after setting to confirm the value was set.')
    parser_driver_set.add_argument('--map-points', help='Take value to be a dictionary of {point: value} mappings.')
    parser_driver_set.set_defaults(func=driver_set_func)

    parser_driver_last = driver_subparsers.add_parser('last', help='Query last known values and/or the corresponding'
                                                                   ' datetime of points on equipment.')
    parser_driver_last.add_argument('topic', help='Topic or pattern to select equipment.').completer = topic_completer
    parser_driver_last.add_argument('-r', '--regex', help='Regex pattern to modify equipment selection.')
    parser_driver_last.add_argument('-t', '--tag', help='Tagging expression to select equipment')
    parser_driver_last.add_argument('-v', '--value', default=True, help='Should the value be returned?')
    parser_driver_last.add_argument('-u', '--updated', default=True, help='Should the datetime of the last known value be returned?')
    parser_driver_last.set_defaults(func=driver_last_func)

    parser_driver_revert = driver_subparsers.add_parser('revert', help='Reset any commands set on equipment.')
    parser_driver_revert.add_argument('topic', help='Topic or pattern to select equipment.').completer = topic_completer
    parser_driver_revert.add_argument('-r', '--regex', help='Regex pattern to modify equipment selection.')
    parser_driver_revert.add_argument('-t', '--tag', help='Tagging expression to select equipment')
    parser_driver_revert.add_argument('-c', '--confirm-values', default=False, help='Query points after reversion to confirm values were reset.')
    parser_driver_revert.set_defaults(func=driver_revert_func)

    parser_driver_start = driver_subparsers.add_parser('start', help='Start polling for a given topic.')
    parser_driver_start.add_argument('topic', help='Topic or pattern to select equipment.').completer = topic_completer
    parser_driver_start.add_argument('-r', '--regex', help='Regex pattern to modify equipment selection.')
    parser_driver_start.add_argument('-t', '--tag', help='Tagging expression to select equipment')
    parser_driver_start.set_defaults(func=driver_start_func)

    parser_driver_stop = driver_subparsers.add_parser('stop', help='Stop polling for a given topic.')
    parser_driver_stop.add_argument('topic', help='Topic or pattern to select equipment.').completer = topic_completer
    parser_driver_stop.add_argument('-r', '--regex', help='Regex pattern to modify equipment selection.')
    parser_driver_stop.add_argument('-t', '--tag', help='Tagging expression to select equipment')
    parser_driver_stop.set_defaults(func=driver_stop_func)

    parser_driver_enable = driver_subparsers.add_parser('enable', help='Set polling to start automatically for topic.')
    parser_driver_enable.add_argument('topic', help='Topic or pattern to select equipment.').completer = topic_completer
    parser_driver_enable.add_argument('-r', '--regex', help='Regex pattern to modify equipment selection.')
    parser_driver_enable.add_argument('-t', '--tag', help='Tagging expression to select equipment')
    parser_driver_enable.set_defaults(func=driver_enable_func)

    parser_driver_disable = driver_subparsers.add_parser('disable', help='Set polling not to start automatically for topic.')
    parser_driver_disable.add_argument('topic', help='Topic or pattern to select equipment.').completer = topic_completer
    parser_driver_disable.add_argument('-r', '--regex', help='Regex pattern to modify equipment selection.')
    parser_driver_disable.add_argument('-t', '--tag', help='Tagging expression to select equipment')
    parser_driver_disable.set_defaults(func=driver_disable_func)

    parser_driver_status = driver_subparsers.add_parser('status', help='Provide status of running drivers.')
    parser_driver_status.add_argument('topic', help='Topic or pattern to select equipment.').completer = topic_completer
    parser_driver_status.add_argument('-r', '--regex', help='Regex pattern to modify equipment selection.')
    parser_driver_status.add_argument('-t', '--tag', help='Tagging expression to select equipment')
    parser_driver_status.set_defaults(func=driver_status_func)

    parser_driver_node = driver_subparsers.add_parser('node', help='Add and remove equipment from Equipment Tree.')
    driver_node_subparsers = parser_driver_node.add_subparsers()

    parser_driver_node_add = driver_node_subparsers.add_parser('add', help='Add a node to the Equipment Tree.')
    parser_driver_node_add.add_argument('node_topic', help='Topic for the new equipment node.')
    parser_driver_node_add.add_argument('-c', '--config', help='Configuration for the new equipment node.')
    parser_driver_node_add.add_argument('-p', '--config-path', help='Path of configuration file for the new equipment node.')
    parser_driver_node_add.add_argument('-u', '--update-schedule', default=True, help='Should the Poll Scheduler be run after adding the node?')
    parser_driver_node_add.set_defaults(func=driver_node_add_func)

    parser_driver_node_remove = driver_node_subparsers.add_parser('remove', help='Remove a node from the Equipment Tree.')
    parser_driver_node_add.add_argument('topic', help='Topic of the equipment node to remove.')
    parser_driver_node_remove.set_defaults(func=driver_node_remove_func)

    parser_driver_interface = driver_subparsers.add_parser('interface', help='Add and remove interfaces.')
    driver_interface_subparsers = parser_driver_interface.add_subparsers()

    parser_driver_interface_install = driver_interface_subparsers.add_parser('install', help='Install a new driver interface.')
    parser_driver_interface_install.add_argument('name', help='Name of the driver interface to install.')
    parser_driver_interface_install.set_defaults(func=driver_interface_install_func)

    parser_driver_interface_list = driver_interface_subparsers.add_parser('list', help='List installed driver interfaces.')
    parser_driver_interface_list.set_defaults(func=driver_interface_list_func)

    parser_driver_interface_remove = driver_interface_subparsers.add_parser('remove', help='Remove an interface from the driver.')
    parser_driver_interface_remove.add_argument('name', help='Name of the driver interface to uninstall.')
    parser_driver_interface_remove.set_defaults(func=driver_interface_remove_func)

    parser_driver_topics = driver_subparsers.add_parser('list', help='List configured equipment topics.')
    parser_driver_topics.add_argument('topic', default='devices', help='Topic or pattern to select equipment.').completer = topic_completer
    parser_driver_topics.add_argument('-r', '--regex', help='Regex pattern to modify equipment selection.')
    parser_driver_topics.add_argument('-t', '--tag', help='Tagging expression to select equipment')
    parser_driver_topics.add_argument('-a', '--active', default=False, help='Exclude topics which are not actively polled.')
    parser_driver_topics.add_argument('-e', '--enabled', default=False, help='Exclude topics are not enabled for polling.')
    parser_driver_topics.set_defaults(func=driver_list_topics_func)

    argcomplete.autocomplete(parser)
    opts = parser.parse_args()

    if not hasattr(opts, "func"):
        parser.print_help()
        sys.exit(0)

    pprint(opts.func(opts))



# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

import json
import logging
import re

from collections import defaultdict
from datetime import datetime, timedelta
from enum import Enum
from importlib.metadata import distribution, PackageNotFoundError
from os.path import normpath
from pydantic import BaseModel
from typing import Any, Union, Iterable
from treelib import Tree, Node
from treelib.exceptions import DuplicatedNodeIdError, NodeIDAbsentError

try:
    distribution('volttron-core')
    from volttron.client.known_identities import CONFIGURATION_STORE
except PackageNotFoundError:
    from volttron.platform.agent.known_identities import CONFIGURATION_STORE

_log = logging.getLogger(__name__)


class TopicNode(Node):
    def __init__(self, tag=None, identifier=None, expanded=True, data=None, segment_type='TOPIC_SEGMENT', topic=''):
        super(TopicNode, self).__init__(tag, identifier, expanded, data)
        self.data: dict[str, Any] = data if data else {}
        self.data['segment_type'] = segment_type
        self.data['topic'] = topic

    @property
    def segment_type(self):
        return self.data['segment_type']

    @property
    def topic(self):
        return self.data['topic']

    def is_segment(self):
        return True if self.segment_type == 'TOPIC_SEGMENT' else False


class TopicTree(Tree):
    def __init__(self, topic_list=None, root_name='root', node_class=None, *args, **kwargs):
        node_class = node_class if node_class else TopicNode
        super(TopicTree, self).__init__(node_class=node_class, *args, **kwargs)
        if topic_list:
            self._from_topic_list(topic_list, root_name)
        else:
            self.create_node(root_name, root_name).data['segment_type'] = 'TOPIC_ROOT'

    def _from_topic_list(self, topic_list, root_name):
        tops = [t.split('/') for t in topic_list]
        if all([top[0] == root_name for top in tops]):
            [top.pop(0) for top in tops]
        self.create_node(root_name, root_name).data['segment_type'] = 'TOPIC_ROOT'
        for top in tops:
            parent = root_name
            for segment in top:
                nid = '/'.join([parent, segment])
                try:
                    self.create_node(segment, nid, parent=parent)
                except DuplicatedNodeIdError:
                    pass
                parent = nid

    def add_node(self, node, parent=None):
        super(TopicTree, self).add_node(node, parent)
        node.data['topic'] = node.identifier[(len(self.root) + 1):]
        return node

    # TODO: Should this actually be get_child_topics() where topics or routes are returned with wildcards?
    def get_children_dict(self, sub_root_node_id: Union[list, str], include_root: bool = True,
                          prefix: str = '', replace_topic: str = None) -> dict:
        sub_root_node_id = sub_root_node_id if type(sub_root_node_id) is list else [sub_root_node_id]
        level_dict = defaultdict(set)
        for r_id in sub_root_node_id:
            for d in self.children(r_id):
                try:
                    if replace_topic:
                        if include_root:
                            level_dict[d.tag].add('/'.join([self.root, replace_topic, d.tag]))
                        else:
                            level_dict[d.tag].add('/'.join([replace_topic, d.tag]))
                    else:
                        if include_root:
                            level_dict[d.tag].add(d.identifier)
                        else:
                            level_dict[d.tag].add(d.identifier.split('/', 1)[1])
                except NodeIDAbsentError as e:
                    return {}
        ret_dict = {}
        for k, s in level_dict.items():
            if len(s) > 1:
                ret_dict[k] = sorted([normpath('/'.join([prefix, v])) for v in s])
            else:
                ret_dict[k] = normpath('/'.join([prefix, s.pop()]))
        return ret_dict

    def resolve_query(self, topic_pattern: str = '', regex: str = None, exact_matches: Iterable = None,
                      return_leaves=False) -> Iterable:
        def clipping(topic_parts, nids=None):
            nids = nids if nids else []
            if topic_parts:
                part = topic_parts.pop(0)
                if part:
                    if not nids:
                        return clipping(topic_parts, [part]) if self.get_node(part) else nids
                    else:
                        joined_nids = [self.get_node('/'.join([c.identifier, part])).identifier
                                       for n in nids for c in self.children(n)]
                        return clipping(topic_parts, joined_nids)
                else:
                    return nids
            else:
                return [l.identifier for n in nids for l in self.leaves(n)] if return_leaves else nids

        if not topic_pattern:
            # If None or empty string, default to "self.root"
            topic_pattern = self.root
        elif not topic_pattern.startswith(self.root):
            topic_pattern = '/'.join([self.root, topic_pattern])

        regex = re.compile(regex) if regex else None
        topic_nids = clipping([part.strip('/') for part in topic_pattern.split('/-') if part != '']) if topic_pattern else []
        if topic_nids and exact_matches:
            nodes = (self.get_node(n) for n in topic_nids if n in exact_matches and (not regex or regex.search(n)))
        elif topic_nids and not exact_matches:
            nodes = (self.get_node(n)  for n in topic_nids if not regex or regex.search(n))
        elif not topic_nids and exact_matches:
            nodes = (self.get_node(n) for n in exact_matches if (not regex or regex.search(n)) and self.contains(n))
        else:
            nodes = self.filter_nodes(lambda n: regex.search(n.identifier)) if regex else []
        return nodes

    def prune(self, topic_pattern: str = None, regex: str = None, exact_matches: Iterable = None, *args, **kwargs):
        if topic_pattern:
            pattern = re.compile(topic_pattern.replace('-', '[^/]+') + '(/|$)')
            nids = [n.identifier for n in self.filter_nodes(lambda x: pattern.search(x.identifier))]
        else:
            nids = list(self.expand_tree())
        if regex:
            regex = re.compile(regex)
            nids = [n for n in nids if regex.search(n)]
        if exact_matches:
            nids = [n for n in nids if n in exact_matches]
        pruned = self.__class__(topic_list=nids, root_name=self.root, *args, **kwargs)
        for nid in [n.identifier for n in pruned.all_nodes()]:
            old = self.get_node(nid)
            pruned.update_node(nid, data=old.data)
        return pruned

    def get_matches(self, topic, return_nodes=True):
        pattern = topic.replace('-', '[^/]+') + '$'
        nodes = self.filter_nodes(lambda x: re.match(pattern, x.identifier))
        if return_nodes:
            return list(nodes)
        else:
            return [n.identifier for n in nodes]

    def to_json(self, with_data=False, sort=True, reverse=False):
        """To format the tree in JSON format."""
        def custom_encoder(obj):
            if isinstance(obj, BaseModel):
                return obj.model_dump()
            if isinstance(obj, Enum):
                return obj.value
            if isinstance(obj, timedelta):
                return obj.total_seconds()
            if isinstance(obj, datetime):
                return obj.isoformat()
            return json.JSONEncoder().default(obj)

        return json.dumps(self.to_dict(with_data=with_data, sort=sort, reverse=reverse), default=custom_encoder)


class DeviceNode(TopicNode):
    def __init__(self, tag=None, identifier=None, expanded=True, data=None, segment_type='TOPIC_SEGMENT', topic=''):
        super(DeviceNode, self).__init__(tag, identifier, expanded, data, segment_type, topic)

    def is_point(self):
        return True if self.segment_type == 'POINT' else False

    def is_device(self):
        return True if self.segment_type == 'DEVICE' else False


class DeviceTree(TopicTree):
    def __init__(self, topic_list=None, root_name='devices', assume_full_topics=False,  *args, **kwargs):
        super(DeviceTree, self).__init__(topic_list=topic_list, root_name=root_name, node_class=DeviceNode,
                                         *args, **kwargs)
        if assume_full_topics:
            for n in self.leaves():
                n.data['segment_type'] = 'POINT'
            for n in [self.parent(l.identifier) for l in self.leaves()]:
                n.data['segment_type'] = 'DEVICE'

    def points(self, nid=None):
        if nid is None:
            points = [n for n in self._nodes.values() if n.is_point()]
        else:
            points = [self[n] for n in self.expand_tree(nid) if self[n].is_point()]
        return points

    def devices(self, nid=None):
        if nid is None:
            points = [n for n in self._nodes.values() if n.is_device()]
        else:
            points = [self[n] for n in self.expand_tree(nid) if self[n].is_device()]
        return points

    # TODO: Getting points requires getting device config, using it to find the registry config,
    #  and then parsing that. There is not a method in config.store, nor in the platform.driver for
    #  getting a completed configuration. The configuration is only fully assembled in the subsystem's
    #  _initial_update method called when the agent itself calls get_configs at startup. There does not
    #  seem to be an equivalent management method, and the code for this is in the agent subsystem
    #  rather than the service (though it is reached through the service, oddly...
    @classmethod
    def from_store(cls, platform, rpc_caller):
        # TODO: Duplicate logic for external_platform check from VUIEndpoints to remove reference to it from here.
        kwargs = {'external_platform': platform} if 'VUIEndpoints' in rpc_caller.__repr__() else {}
        devices = rpc_caller(CONFIGURATION_STORE, 'manage_list_configs', 'platform.driver', **kwargs)
        devices = devices if kwargs else devices.get(timeout=5)
        devices = [d for d in devices if re.match('^devices/.*', d)]
        device_tree = cls(devices)
        for d in devices:
            dev_config = rpc_caller(CONFIGURATION_STORE, 'manage_get', 'platform.driver', d, raw=False, **kwargs)
            # TODO: If not AsyncResponse instead of if kwargs
            dev_config = dev_config if kwargs else dev_config.get(timeout=5)
            reg_cfg_name = dev_config.pop('registry_config')[len('config://'):]
            data = {'config': dev_config, 'segment_type': 'DEVICE'}
            device_tree.update_node(d, data=data)
            registry_config = rpc_caller('config.store', 'manage_get', 'platform.driver',
                                         f'{reg_cfg_name}', raw=False, **kwargs)
            registry_config = registry_config if kwargs else registry_config.get(timeout=5)
            for pnt in registry_config:
                point_name = pnt.pop('Volttron Point Name')
                n = device_tree.create_node(point_name, f"{d}/{point_name}", parent=d, data=pnt)
                n.data['segment_type'] = 'POINT'
        return device_tree



# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

import fnmatch
import logging

from datetime import datetime, timedelta
from importlib.metadata import distribution, PackageNotFoundError

try:
    distribution('volttron-core')
    from volttron.client.vip.agent.core import ScheduledEvent
    from volttron.driver.base.interfaces import DriverInterfaceError
    from volttron.utils import format_timestamp, get_aware_utc_now, parse_timestamp_string
    from volttron.utils.jsonapi import dumps, loads
except PackageNotFoundError:
    from volttron.platform.vip.agent.core import ScheduledEvent
    from volttron.driver.base.interfaces import DriverInterfaceError
    from volttron.platform.agent.utils import format_timestamp, get_aware_utc_now, parse_timestamp_string
    from volttron.platform.jsonapi import dumps, loads

_log = logging.getLogger(__name__)


class OverrideError(DriverInterfaceError):
    """Error raised when the user tries to set/revert point when global override is set."""
    pass

# TODO: Rework the logic in this class to use new data structures instead of self.instances.
class OverrideManager:
    def __init__(self, parent):
        self.devices = set()
        self.interval_events: dict[str, tuple[ScheduledEvent, datetime] | None]  = {}
        self.agent = parent
        self.patterns = set()

        try:
            values = loads(self.agent.vip.config.get("_override_patterns"))
            if isinstance(values, dict):
                for pattern, end_time in values.items():
                    now = get_aware_utc_now()
                    if end_time == "0.0":   # If end time is indefinite, set override with indefinite duration
                        self.set_on(pattern, 0.0, from_config_store=True)
                    else:
                        end_time = parse_timestamp_string(end_time)
                        if end_time > now:  # If end time > current time, set override with new duration
                            self.set_on(pattern, (end_time - now).total_seconds(), from_config_store=True)
        except KeyError:
            self.patterns = set()
        except ValueError:
            _log.error("Override patterns is not set correctly in config store")
            self.patterns = set()

    def set_on(self,
               pattern,
               duration=0.0,
               failsafe_revert=True,
               staggered_revert=False,
               from_config_store=False):
        """
        Turn on override condition on all devices matching the pattern. It schedules
        an event to keep track of the duration over which override has to be applied.

        :param pattern: Override pattern (wildcard) to be applied, e.g. "devices/some_device/*".
        :param duration: Time duration for the override in seconds. If <= 0.0, it's indefinite.
        :param failsafe_revert: If True, revert points to their default state.
        :param staggered_revert: If True, revert calls are staggered over time.
        :param from_config_store: If True, indicates this method was called from a config store callback.
        """
        _log.debug(f"Setting override on pattern='{pattern}', duration={duration}, "
                   f"failsafe_revert={failsafe_revert}, staggered_revert={staggered_revert}")

        stagger_interval = 0.05  # seconds between staggered reverts

        # Add this pattern to our set of known override patterns
        self.patterns.add(pattern)

        i = 0
        # Loop over each device from the equipment tree
        # 'devices()' might return strings or DeviceNodes
        for device_obj in self.agent.equipment_tree.devices():
            i += 1

            # Convert DeviceNode -> string, if needed
            if hasattr(device_obj, "identifier"):
                device_name = device_obj.identifier  # e.g. "devices/singletestfake"
            else:
                device_name = device_obj  # it might already be a string

            # Match the device_name against the override pattern
            if fnmatch.fnmatch(device_name, pattern):
                _log.debug(f"Override matched device='{device_name}' with pattern='{pattern}'")

                # If failsafe_revert is requested, revert the device
                if failsafe_revert:
                    if staggered_revert:
                        # Stagger this revert call
                        self.agent.core.spawn_later(
                            i * stagger_interval, self.agent.revert, device_name
                        )
                    else:
                        # Immediately revert
                        self.agent.core.spawn(self.agent.revert, device_name)

                # Mark this device as overridden
                self.devices.add(device_name)

        # Set a timer for the override duration (if > 0)
        config_update = self._update_override_interval(duration, pattern)

        # If we changed the override intervals, and it wasn't a config store callback,
        # then update our stored override patterns in config store
        if config_update and not from_config_store:
            patterns_dict = {}
            for pat in self.patterns:
                if self.interval_events.get(pat) is None:
                    patterns_dict[pat] = "0.0"
                else:
                    evt, end_time = self.interval_events[pat]
                    patterns_dict[pat] = format_timestamp(end_time)

            self.agent.vip.config.set("_override_patterns", dumps(patterns_dict))

    def set_off(self, pattern):
        """Turn off override condition on all devices matching the pattern. It removes the pattern from the override
        patterns set, clears the list of overridden devices  and reevaluates the state of devices. It then cancels the
        pending override event and removes pattern from the config store.
        :param pattern: Override pattern to be removed.
        :type pattern: str
        """
        # If pattern exactly matches
        if pattern in self.patterns:
            self.patterns.discard(pattern)
            # Cancel any pending override events
            self._cancel_override_events(pattern)
            self.devices.clear()
            patterns = dict()
            # Build override devices list again
            for pat in self.patterns:
                for device in self.agent.equipment_tree.devices():
                    if fnmatch.fnmatch(device, pat):
                        self.devices.add(device)

                if self.interval_events[pat] is None:
                    patterns[pat] = str(0.0)
                else:
                    evt, end_time = self.interval_events[pat]
                    patterns[pat] = format_timestamp(end_time)

            self.agent.vip.config.set("_override_patterns", dumps(patterns))
        else:
            _log.error("Override Pattern did not match!")
            raise OverrideError(
                "Pattern {} does not exist in list of override patterns".format(pattern))

    def clear(self):
        """RPC method

        Clear all overrides.
        """
        # Cancel all pending override timer events
        for pattern, evt in self.interval_events.items():
            if evt is not None:
                evt[0].cancel()
        self.interval_events.clear()
        self.devices.clear()
        self.patterns.clear()
        self.agent.vip.config.set("_override_patterns", {})

    def _update_override_interval(self, interval, pattern):
        """Schedules a new override event for the specified interval and pattern. If the pattern already exists and new
        end time is greater than old one, the event is cancelled and new event is scheduled.

        :param interval: override duration. If interval is <= 0.0, implies indefinite duration
        :type pattern: float
        :param pattern: Override pattern.
        :type pattern: str
        :return Flag to indicate if update is done or not.
        """
        if interval <= 0.0:    # indicative of indefinite duration
            if pattern in self.interval_events:
                # If override duration is indefinite, do nothing
                if self.interval_events[pattern] is None:
                    return False
                else:
                    # Cancel the old event
                    evt = self.interval_events.pop(pattern)
                    evt[0].cancel()
            self.interval_events[pattern] = None
            return True
        else:
            override_start = get_aware_utc_now()
            override_end = override_start + timedelta(seconds=interval)
            if pattern in self.interval_events:
                evt = self.interval_events[pattern]
                # If event is indefinite or greater than new end time, do nothing
                if evt is None or override_end < evt[1]:
                    return False
                else:
                    evt = self.interval_events.pop(pattern)
                    evt[0].cancel()
            # Schedule new override event
            event = self.agent.core.schedule(override_end, self.set_off, pattern)
            self.interval_events[pattern] = (event, override_end)
            return True

    def _cancel_override_events(self, pattern):
        """
        Cancel override event matching the pattern
        :param pattern: override pattern
        :type pattern: str
        """
        if pattern in self.interval_events:
            # Cancel the override cancellation timer event
            evt = self.interval_events.pop(pattern, None)
            if evt is not None:
                evt[0].cancel()

    def _update_override_state(self, device, state):
        """
        If a new device is added, it is checked to see if the device is part of the list of overridden patterns. If so,
        it is added to the list of overridden devices. Similarly, if a device is being removed, it is also removed
        from list of overridden devices (if exists).
        :param device: device to be removed
        :type device: str
        :param state: 'add' or 'remove'
        :type state: str
        """
        device = device.lower()

        if state == 'add':
            # If device falls under the existing overridden patterns, then add it to list of overridden devices.
            for pattern in self.patterns:
                if fnmatch.fnmatch(device, pattern):
                    self.devices.add(device)
                    return
        else:
            # If device is in list of overridden devices, remove it.
            if device in self.devices:
                self.devices.remove(device)



# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

import bisect
import logging
import pickle

from base64 import b64encode
from collections import defaultdict, namedtuple
from copy import deepcopy
from datetime import timedelta
from importlib.metadata import distribution, PackageNotFoundError
from pickle import dumps, loads
from tzlocal import get_localzone

try:
    distribution('volttron-core')
    from volttron.client.messaging import topics
    from volttron.utils import get_aware_utc_now, format_timestamp, parse_timestamp_string
except PackageNotFoundError:
    from volttron.platform.messaging import topics
    from volttron.platform.agent.utils import get_aware_utc_now, format_timestamp, parse_timestamp_string

PRIORITY_HIGH = 'HIGH'
PRIORITY_LOW = 'LOW'
PRIORITY_LOW_PREEMPT = 'LOW_PREEMPT'
ALL_PRIORITIES = {PRIORITY_HIGH, PRIORITY_LOW, PRIORITY_LOW_PREEMPT}

# TODO: Update volttron.client.messaging.topics?
ACTUATOR_RESERVATION_ANNOUNCE_RAW = topics.ACTUATOR_SCHEDULE_ANNOUNCE_RAW

# RequestResult - Result of a reservation request returned from the reservation
# manager.
RequestResult = namedtuple('RequestResult', ['success', 'data', 'info_string'])
DeviceState = namedtuple('DeviceState', ['agent_id', 'task_id', 'time_remaining'])
_log = logging.getLogger(__name__)


class TimeSlice(object):

    def __init__(self, start=None, end=None):
        if end is None:
            end = start
        if start is not None:
            if end < start:
                raise ValueError('Invalid start and end values.')
        self._start = start
        self._end = end

    def __repr__(self):
        return 'TimeSlice({start!r},{end!r})'.format(start=self.start, end=self.end)

    def __str__(self):
        return '({start} <-> {end})'.format(start=self.start, end=self.end)

    @property
    def end(self):
        return self._end

    @property
    def start(self):
        return self._start

    def __cmp__(self, other):
        if self.start >= other.end:
            return 1
        if self.end <= other.start:
            return -1
        return 0

    # def __ne__(self, other):
    #     return self.__cmp__(other) != 0
    #
    # def __gt__(self, other):
    #     return self.__cmp__(other) > 0

    def __lt__(self, other):
        return self.__cmp__(other) < 0

    # def __ge__(self, other):
    #     return self.__cmp__(other) >= 0
    #
    # def __le__(self, other):
    #     return self.__cmp__(other) <= 0

    def __contains__(self, other):
        return self.start < other < self.end

    def stretch_to_include(self, time_slice):
        if self.start is None or time_slice.start < self.start:
            self._start = time_slice.start
        if self.end is None or time_slice.end > self.end:
            self._end = time_slice.end

    def contains_include_start(self, other):
        """Similar to == or "in" but includes time == self.start"""
        return other in self or other == self.start


class Task(object):
    STATE_PRE_RUN = 'PRE_RUN'
    STATE_RUNNING = 'RUNNING'
    STATE_PREEMPTED = 'PREEMPTED'
    STATE_FINISHED = 'FINISHED'

    def __init__(self, agent_id, priority, requests):
        self.agent_id = agent_id
        self.priority = priority
        self.time_slice = TimeSlice()
        self.devices = defaultdict(Reservation)
        self.state = Task.STATE_PRE_RUN
        self.populate_reservation(requests)

    def change_state(self, new_state):
        if self.state == new_state:
            return

        # TODO: We can put code here for managing state changes.

        self.state = new_state

    def populate_reservation(self, requests):
        for request in requests:
            device, start, end = request

            time_slice = TimeSlice(start, end)
            if not isinstance(device, str):
                raise ValueError('Device not string.')
            self.devices[device].reserve_slot(time_slice)
            self.time_slice.stretch_to_include(time_slice)

    def make_current(self, now):
        if self.state == Task.STATE_FINISHED:
            self.devices.clear()
            return

        for device, reservation in list(self.devices.items()):
            if reservation.finished(now):
                del self.devices[device]

        if self.time_slice.contains_include_start(now):
            if self.state != Task.STATE_PREEMPTED:
                self.change_state(Task.STATE_RUNNING)

        elif self.time_slice > TimeSlice(now):
            self.change_state(Task.STATE_PRE_RUN)

        elif self.time_slice < TimeSlice(now):
            self.change_state(Task.STATE_FINISHED)

    def get_current_slots(self, now):
        result = {}
        for device, reservation in self.devices.items():
            time_slot = reservation.get_current_slot(now)
            if time_slot is not None:
                result[device] = time_slot

        return result

    def get_conflicts(self, other):
        results = []
        for device, reservation in self.devices.items():
            if device in other.devices:
                conflicts = other.devices[device].get_conflicts(reservation)
                results.extend([device, str(x.start), str(x.end)] for x in conflicts)

        return results

    def check_can_preempt_other(self, other):
        if self.priority != PRIORITY_HIGH:
            return False

        if other.priority == PRIORITY_HIGH:
            return False

        if other.state == Task.STATE_RUNNING and other.priority != \
                PRIORITY_LOW_PREEMPT:
            return False

        return True

    def preempt(self, grace_time, now):
        """Return true if there are time slots that have a grace period left"""
        self.make_current(now)
        if self.state == Task.STATE_PREEMPTED:
            return True
        if self.state == Task.STATE_FINISHED:
            return False

        current_time_slots = []
        for reservation in self.devices.values():
            current_time_slots.extend(reservation.prune_to_current(grace_time, now))

        self.change_state(Task.STATE_FINISHED if not current_time_slots else Task.STATE_PREEMPTED)

        if self.state == Task.STATE_PREEMPTED:
            self.time_slice = TimeSlice(now, now + grace_time)
            return True

        return False

    def get_next_event_time(self, now):
        device_reservations = (x.get_next_event_time(now) for x in self.devices.values())
        events = [x for x in device_reservations if x is not None]

        if events:
            return min(events)

        return None

class LockError(Exception):
    # Superclass of ReservationLockError for backwards compatability to old Actuator Agent.
    pass

class ReservationLockError(LockError):
    """Error raised when the user does not have a device scheduled
    and tries to use methods that require exclusive access."""
    pass

class ReservationSchedulingError(Exception):
    pass


class Reservation(object):

    def __init__(self):
        self.time_slots = []

    def check_availability(self, time_slot):
        start_slice = bisect.bisect_left(self.time_slots, time_slot)
        end_slice = bisect.bisect_right(self.time_slots, time_slot)
        return set(self.time_slots[start_slice:end_slice])

    def make_current(self, now):
        """Should be called before working with a reservation.
        Updates the state to the reservation to eliminate stuff in the past."""
        now_slice = bisect.bisect_left(self.time_slots, TimeSlice(now))
        _log.debug("now_slice in make_current {}".format(now_slice))
        if now_slice > 0:
            del self.time_slots[:now_slice]

    def reserve_slot(self, time_slot):
        if self.check_availability(time_slot):
            raise ReservationSchedulingError('DERP! We messed up the scheduling!')

        bisect.insort(self.time_slots, time_slot)

    def get_next_event_time(self, now):
        """Run this to know when to the next state change is going to happen
        with this reservation"""
        self.make_current(now)
        if not self.time_slots:
            return None
        _log.debug(f"in reservation get_next_event_time timeslots {self.time_slots[0]} now {now}")
        next_time = self.time_slots[0].end if self.time_slots[0].contains_include_start(
            now) else self.time_slots[0].start
        # Round to the next second to fix timer goofiness in agent timers.
        if next_time.microsecond:
            next_time = next_time.replace(microsecond=0) + timedelta(seconds=1)

        return next_time

    def get_current_slot(self, now):
        """
        Determines if "now" falls within any scheduled time slots and returns the current active slot, or None if no slot is active
        """
        self.make_current(now)
        if not self.time_slots:
            return None

        if self.time_slots[0].contains_include_start(now):
            return self.time_slots[0]

        return None

    def prune_to_current(self, grace_time, now):
        """Use this to prune a reservation due to preemption."""
        current_slot = self.get_current_slot(now)
        if current_slot is not None:
            latest_end = now + grace_time
            if current_slot.end > latest_end:
                current_slot = TimeSlice(current_slot.start, latest_end)
            self.time_slots = [current_slot]
        else:
            self.time_slots = []

        return self.time_slots

    def get_conflicts(self, other):
        """Returns a list of our time_slices that conflict with the other
        reservation"""
        return [x for x in self.time_slots if other.check_availability(x)]

    def finished(self, now):
        self.make_current(now)
        return not bool(self.time_slots)

    def get_reservation(self):
        return deepcopy(self.time_slots)

    def __len__(self):
        return len(self.time_slots)

    def __repr__(self):
        pass


class ReservationManager(object):

    def __init__(self, parent, grace_time, now=None):
        self.agent = parent
        self.tasks = {}
        self.running_tasks = set()
        self.preempted_tasks = set()
        self.grace_time = timedelta(grace_time)

        self._device_states = {}
        self.reservation_state_file = "_reservation_state"
        self._update_event = None
        self._update_event_time = None

        try:
            initial_state_string = self.agent.vip.config.get(self.reservation_state_file)
        except KeyError:
            initial_state_string = None
        now = now if now else get_aware_utc_now()
        self.load_state(now, initial_state_string)

    def reserved_by(self, topic):
        return self._device_states.get(topic)

    def update(self, now, device_only=None, publish=True):
        # Sanity check now.
        # This is specifically for when this is running in a VM that gets
        # suspended and then resumed.
        # If we don't make this check a resumed VM will publish one event
        # per minute of
        # time the VM was suspended for.

        test_now = get_aware_utc_now()
        if test_now - now > timedelta(minutes=3):
            now = test_now
        self._device_states = self.get_reservation_state(now)

        # device_only and publish tells us if we were called by a reservation change.
        # If we are being called as part of a regularly scheduled publish
        # we ignore our previous publication time.
        if device_only is None and publish:
            self._update_event_time = None

        next_reservation_event_time = self.get_next_event_time(now)
        new_update_event_time = self._get_adjusted_next_event_time(now, next_reservation_event_time,
                                                                   self._update_event_time)

        if publish:
            device_states = []
            if device_only is not None:
                # TODO: What if there are multiple devices in a newly scheduled task? Seems like this should loop.
                if device_only in self._device_states:
                    device_states.append((device_only, self._device_states[device_only]))
            else:
                device_states = iter(self._device_states.items())

            for device, state in device_states:
                _log.debug("device, state -  {}, {}".format(device, state))
                headers = {
                    'time': format_timestamp(now),
                    'requesterID': state.agent_id,
                    'taskID': state.task_id,
                    'window': state.time_remaining
                }
                topic = ACTUATOR_RESERVATION_ANNOUNCE_RAW.replace('{device}', device)
                self.agent.vip.pubsub.publish('pubsub', topic, headers=headers)

        if self._update_event is not None:
            # This won't hurt anything if we are canceling ourselves.
            self._update_event.cancel()
        self._update_event_time = new_update_event_time
        self._update_event = self.agent.core.schedule(new_update_event_time, self.update,
                                                      new_update_event_time)

    # # TODO: Is this function necessary?
    # def _update_reservation_state(self, now):
    #     self.update(now)

    def _get_adjusted_next_event_time(self, now, next_event_time, previously_reserved_time):
        latest_next = now + timedelta(seconds=self.agent.config.reservation_publish_interval)
        # Round to the next second to fix timer goofiness in agent timers.
        # TODO: Improved Reservation Manager should no longer require this.
        if latest_next.microsecond:
            latest_next = latest_next.replace(microsecond=0) + timedelta(seconds=1)

        result = latest_next
        if next_event_time is not None and result > next_event_time:
            result = next_event_time

        if previously_reserved_time is not None and result > previously_reserved_time:
            result = previously_reserved_time

        return result

    def set_grace_period(self, seconds):
        self.grace_time = timedelta(seconds=seconds)

    def load_state(self, now, initial_state_string):
        if initial_state_string is None:
            return

        try:
            self.tasks = loads(initial_state_string)
            self._cleanup(now)
        except pickle.PickleError as pe:
            self.tasks = {}
            _log.error(f'Pickle error {pe}')
        except Exception as e:
            self.tasks = {}
            _log.error(f'Reservation Manager state file corrupted! Exception {e}')

    def save_state(self, now):
        try:
            self._cleanup(now)
            _log.debug(f"Saving {len(self.tasks)} task(s)")
            self.agent.vip.config.set(self.reservation_state_file, b64encode(dumps(self.tasks)).decode("utf-8"), send_update=False)

        except Exception as e:
            _log.error(f'Failed to save Reservation Manager state! Error: {e}')

    def new_task(self, sender, task_id, priority, requests, now=None):
        priority = priority.upper() if priority is not None else None
        local_tz = get_localzone()
        if requests and isinstance(requests[0], str):
            requests = [requests]

        tmp_requests = requests
        requests = []
        for r in tmp_requests:
            device, start, end = r

            device = device.strip('/')
            start = parse_timestamp_string(start)
            end = parse_timestamp_string(end)
            if start.tzinfo is None:
                start = local_tz.localize(start)
            if end.tzinfo is None:
                end = local_tz.localize(end)
            requests.append([device, start, end])
        _log.debug(f"Got new reservation request: {sender}, {task_id}, {priority}, {requests}")
        
        now = now if now is not None else get_aware_utc_now()
        self._cleanup(now)

        if task_id in self.tasks:
            return RequestResult(False, {}, 'TASK_ID_ALREADY_EXISTS')
        if task_id is None:
            return RequestResult(False, {}, 'MISSING_TASK_ID')
        if priority is None:
            return RequestResult(False, {}, 'MISSING_PRIORITY')
        if priority not in ALL_PRIORITIES:
            return RequestResult(False, {}, 'INVALID_PRIORITY')
        if sender is None:
            return RequestResult(False, {}, 'MISSING_AGENT_ID')
        if requests is None or not requests:
            return RequestResult(False, {}, 'MALFORMED_REQUEST_EMPTY')
        if not isinstance(sender, str) or not sender:
            return RequestResult(
                False, {}, 'MALFORMED_REQUEST: TypeError: agent_id must '
                'be a nonempty string')
        if not isinstance(task_id, str) or not task_id:
            return RequestResult(
                False, {}, 'MALFORMED_REQUEST: TypeError: taskid must '
                'be a nonempty string')

        try:
            new_task = Task(sender, priority, requests)
        except ReservationSchedulingError:
            return RequestResult(False, {}, 'REQUEST_CONFLICTS_WITH_SELF')
        except Exception as ex:
            return RequestResult(False, {},
                                 'MALFORMED_REQUEST: ' + ex.__class__.__name__ + ': ' + str(ex))

        conflicts = defaultdict(dict)
        preempted_tasks = set()

        for t_id, task in self.tasks.items():
            conflict_list = new_task.get_conflicts(task)
            sender = task.agent_id
            if conflict_list:
                if not new_task.check_can_preempt_other(task):
                    conflicts[sender][t_id] = conflict_list
                else:
                    preempted_tasks.add((sender, t_id))
        if conflicts:
            return RequestResult(False, conflicts, 'CONFLICTS_WITH_EXISTING_RESERVATIONS')
        self.tasks[task_id] = new_task
        _log.debug(f"Task added. Total tasks now: {len(self.tasks)}")

        # By this point we know that any remaining conflicts can be preempted and the request will succeed.
        for _, t_id in preempted_tasks:
            task = self.tasks[t_id]
            task.preempt(self.grace_time, now)
        self.save_state(now)

        if preempted_tasks:
            result = RequestResult(True, list(preempted_tasks), 'TASKS_WERE_PREEMPTED')
        else:
            return RequestResult(True, {}, '')
        if result.success:
            # TODO: This implies a single device, but the loop above implies potentially multiple.
            #  The update() method sends a publish on the announce/full/device/path topic, implying all devices.
            #  Neither here nor update() actually have logic for multiple devices, though.
            #  This is a carryover bug from the Actuator Agent. Investigate the format of the request.
            self.update(now, device_only=device)
        return result

    def cancel_task(self, sender, task_id):
        now = get_aware_utc_now()
        if task_id not in self.tasks:
            return RequestResult(False, {}, 'TASK_ID_DOES_NOT_EXIST')
        task = self.tasks[task_id]
        if task.agent_id != sender:
            return RequestResult(False, {}, 'AGENT_ID_TASK_ID_MISMATCH')
        del self.tasks[task_id]
        _log.debug(f"Task {task_id} successfully cancelled.")
        self.save_state(now)
        result = RequestResult(True, {}, '')
        if result.success:
            self.update(now, publish=False)
        return result

    def get_reservation_state(self, now):
        self._cleanup(now)
        running_results = {}
        preempted_results = {}
        for task_id in self.running_tasks:
            task = self.tasks[task_id]
            agent_id = task.agent_id
            current_task_slots = task.get_current_slots(now)
            _log.debug("current_task_slots {}".format(current_task_slots))
            for device, time_slot in current_task_slots.items():
                assert (device not in running_results)
                running_results[device] = DeviceState(agent_id, task_id,
                                                      (time_slot.end - now).total_seconds())

        for task_id in self.preempted_tasks:
            task = self.tasks[task_id]
            agent_id = task.agent_id
            current_task_slots = task.get_current_slots(now)
            for device, time_slot in current_task_slots.items():
                assert (device not in preempted_results)
                preempted_results[device] = DeviceState(agent_id, task_id,
                                                        (time_slot.end - now).total_seconds())

        running_results.update(preempted_results)
        return running_results

    def get_next_event_time(self, now):
        task_times = (x.get_next_event_time(now) for x in self.tasks.values())
        events = [x for x in task_times if x is not None]

        if events:
            return min(events)

        return None

    def _cleanup(self, now):
        """Cleans up self and contained tasks to reflect the current time.
        Should be called:
        1. Before serializing to disk.
        2. After reading from disk.
        3. Before handling a reservation submission request.
        4. After handling a reservation submission request.
        5. Before handling a state request."""
        # Reset the running tasks.
        self.running_tasks = set()
        self.preempted_tasks = set()

        for task_id in list(self.tasks.keys()):
            task = self.tasks[task_id]
            task.make_current(now)
            if task.state == Task.STATE_FINISHED:
                _log.debug(f"Removing task '{task_id}' because it is finished.")
                del self.tasks[task_id]

            elif task.state == Task.STATE_RUNNING:
                self.running_tasks.add(task_id)

            elif task.state == Task.STATE_PREEMPTED:
                self.preempted_tasks.add(task_id)

    def __repr__(self):
        pass



# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

import abc
import gevent
import importlib  # TODO: Look into using "get_module", "get_class", "get_subclasses" from volttron.utils.dynamic_helper
import logging

from collections import defaultdict
from datetime import datetime, timedelta
from importlib.metadata import distribution, PackageNotFoundError
from functools import reduce
from math import floor, gcd, lcm
from typing import Any
from weakref import WeakKeyDictionary, WeakValueDictionary

try:
    distribution('volttron-core')
    # noinspection PyProtectedMember
    from volttron.client.vip.agent.core import ScheduledEvent
    from volttron.driver.base.driver import DriverAgent
    from volttron.utils import get_aware_utc_now
except PackageNotFoundError:
    # noinspection PyProtectedMember
    from volttron.platform.vip.agent.core import ScheduledEvent
    from volttron.driver.base.driver import DriverAgent
    from volttron.platform.agent.utils import get_aware_utc_now

from .config import GroupConfig
from .equipment import EquipmentTree, PointNode

_log = logging.getLogger(__name__)


class PollSet:
    def __init__(self, data_model: EquipmentTree, remote: DriverAgent,
                 points: WeakValueDictionary[str, PointNode] = None,
                 single_depth: set[str] = None, single_breadth: set[tuple[str, str]] = None,
                 multi_depth: dict[str, set[str]] = None, multi_breadth: dict[str, set[str]] = None):
        self.data_model: EquipmentTree = data_model
        self.points: WeakValueDictionary[str, PointNode] = points if points else WeakValueDictionary()
        self.single_depth: set[str] = single_depth if single_depth else set()
        self.single_breadth: set[tuple[str, str]] = single_breadth if single_breadth else set()
        self.multi_depth: dict[str, set[str]] = multi_depth if multi_depth else defaultdict(set)
        self.multi_breadth: dict[str, set[str]] = multi_breadth if multi_breadth else defaultdict(set)
        self.remote = remote

    def add(self, item: PointNode):
        self.points[item.identifier] = item
        self._add_to_publish_setup(item)

    def remove(self, item: PointNode):
        success = self.points.pop(item.identifier, None)
        self._remove_from_publish_setup(item)
        return success

    def _add_to_publish_setup(self, point: PointNode):
        point_depth, point_breadth = self.data_model.get_point_topics(point.identifier)
        device_depth, device_breadth = self.data_model.get_device_topics(point.identifier)

        if self.data_model.is_published_single_depth(point.identifier):
            self.single_depth.add(point_depth)

        if self.data_model.is_published_single_breadth(point.identifier):
            self.single_breadth.add((point_depth, point_breadth))

        if self.data_model.is_published_multi_depth(point.identifier):
            self.multi_depth[device_depth].add(point_depth)

        if self.data_model.is_published_multi_breadth(point.identifier):
            self.multi_breadth[device_breadth].add(point.identifier)

    def  _remove_from_publish_setup(self, point: PointNode):
        point_depth, point_breadth = self.data_model.get_point_topics(point.identifier)
        device_depth, device_breadth = self.data_model.get_device_topics(point.identifier)
        self.single_depth.discard(point_depth)
        self.single_breadth.discard((point_depth, point_breadth))
        self.multi_depth[device_depth].discard(point_depth)
        if not self.multi_depth[device_depth]:
            self.multi_depth.pop(device_depth, None)
        self.multi_breadth[device_breadth].discard(point.identifier)
        if not self.multi_breadth[device_breadth]:
            self.multi_breadth.pop(device_breadth, None)

    def __or__(self, other):
        if self.data_model is not other.data_model:
            raise ValueError(f'Cannot combine PollSets based on different data models:'
                             f' {self.data_model}, {other.data_model}.')
        # TODO: Why can't we have polling schedules for multiple remotes -- e.g., a serial bus with many remotes.
        if self.remote is not other.remote:
            raise ValueError(f'Cannot combine PollSets based on different remotes:'
                             f' {self.remote.unique_id}, {other.remote.unique_id}.')
        return PollSet(
            self.data_model,
            self.remote,
            points=self.points | other.points,
            single_depth=self.single_depth | other.single_depth,
            single_breadth=self.single_breadth | other.single_breadth,
            multi_depth={k: self.multi_depth.get(k, set()) | other.multi_depth.get(k, set())
                    for k in self.multi_depth.keys() | other.multi_depth.keys()},
            multi_breadth={k: self.multi_breadth.get(k, set()) | other.multi_breadth.get(k, set())
                    for k in self.multi_breadth.keys() | other.multi_breadth.keys()}
        )

    def __bool__(self):
        return bool(self.points)


class PollScheduler(metaclass=abc.ABCMeta):
    poll_sets: dict[str, WeakKeyDictionary[DriverAgent, dict[float, PollSet]]] = defaultdict(WeakKeyDictionary)

    def __init__(self, data_model: EquipmentTree, group: str, group_config: GroupConfig, **kwargs):
        self.data_model: EquipmentTree = data_model
        self.group: str = group
        self.group_config: GroupConfig = group_config

        self.start_all_datetime: datetime = get_aware_utc_now()
        self.pollers: dict[Any, ScheduledEvent] = {}

    def schedule(self):
        self._prepare_to_schedule()
        self._schedule_polling()

    @classmethod
    def setup(cls, data_model: EquipmentTree, group_configs: dict[str, GroupConfig]):
        """
        Sort points from each of the remote's EquipmentNodes by interval:
            Build cls.poll_sets  as: {group: {remote: {interval: WeakSet(points)}}}
        """
        cls._build_poll_sets(data_model)
        poll_schedulers = cls.create_poll_schedulers(data_model, group_configs)
        return poll_schedulers

    @classmethod
    def create_poll_schedulers(cls, data_model: EquipmentTree, group_configs,
                               specific_groups: list[str] = None, existing_group_count: int = 0):
        poll_schedulers = {}
        groups = specific_groups if specific_groups else cls.poll_sets
        for i, group in enumerate(groups):
            group_config = group_configs.get(group)
            if group_config is None:
                # Create a config for the group with default settings and mimic the old offset multiplier behavior.
                group_config: GroupConfig = GroupConfig()
                # TODO: Should start_offset instead be either a default offset * i or the specified start_offset if it is there?
                group_config.start_offset = group_config.start_offset * (i + existing_group_count)
                group_configs[group] = group_config  # Add this new config back to the agent settings.
                # TODO: Save the agent settings afterwards so this group gets the same config next time?
            poll_scheduler_module = importlib.import_module(group_config.poll_scheduler_module)
            poll_scheduler_class = getattr(poll_scheduler_module, group_config.poll_scheduler_class_name)
            poll_schedulers[group] = poll_scheduler_class(data_model, group, group_config)
        return poll_schedulers

    @classmethod
    def _build_poll_sets(cls, data_model: EquipmentTree):
        for remote in data_model.remotes.values():
            poll_sets = defaultdict(dict)
            groups = set()
            for point in remote.point_set:
                if data_model.is_active(point.identifier):
                    group = data_model.get_group(point.identifier)
                    interval = data_model.get_polling_interval(point.identifier)
                    if interval not in poll_sets[group]:
                        poll_sets[group][interval] = PollSet(data_model, remote)
                    poll_sets[group][interval].add(point)
                    groups.add(group)
            for group in poll_sets:
                if remote not in cls.poll_sets[group]:
                    cls.poll_sets[group][remote] = defaultdict(lambda: PollSet(data_model, remote))
                for interval in poll_sets[group]:
                    cls.poll_sets[group][remote][interval] = poll_sets[group][interval]

    @staticmethod
    def find_starting_datetime(now: datetime, interval: timedelta, group_delay: timedelta = None):
        group_delay = timedelta(seconds=0.0) if not isinstance(group_delay, timedelta) else group_delay
        midnight = now.replace(hour=0, minute=0, second=0, microsecond=0)
        seconds_from_midnight = (now - midnight)
        offset = seconds_from_midnight % interval
        if not offset:
            return now + interval + group_delay
        next_from_midnight = seconds_from_midnight - offset + interval
        return midnight + next_from_midnight + group_delay

    @classmethod
    def add_to_schedule(cls, point: PointNode, data_model: EquipmentTree):
        """Add a poll to the schedule, without complete rescheduling if possible"""
        group = data_model.get_group(point.identifier)
        remote = data_model.get_remote(point.identifier)
        interval = data_model.get_polling_interval(point.identifier)
        reschedule_required = (group not in cls.poll_sets
                               or remote not in cls.poll_sets[group]
                               or interval not in cls.poll_sets[group][remote])
        if remote not in cls.poll_sets[group].keys():
            cls.poll_sets[group][remote] = defaultdict(lambda: PollSet(data_model, remote))
        cls.poll_sets[group][remote][interval].add(point)
        return reschedule_required

    @classmethod
    def remove_from_schedule(cls, point: PointNode, data_model: EquipmentTree):
        """Remove a poll from the schedule without rescheduling."""
        group = data_model.get_group(point.identifier)
        remote = data_model.get_remote(point.identifier)
        interval = data_model.get_polling_interval(point.identifier)
        success = cls.poll_sets[group][remote][interval].remove(point)
        cls._prune_poll_sets(group, interval, remote)
        return True if success else False

    @classmethod
    def _prune_poll_sets(cls, group, interval, remote):
        if not cls.poll_sets[group][remote][interval].points:
            cls.poll_sets[group][remote].pop(interval)
            if not cls.poll_sets[group][remote]:
                cls.poll_sets[group].pop(remote)
                if not cls.poll_sets[group]:
                    cls.poll_sets.pop(group)

    @abc.abstractmethod
    def _prepare_to_schedule(self):
        pass

    @abc.abstractmethod
    def _schedule_polling(self):
        pass

    @abc.abstractmethod
    def get_schedule(self):
        pass


class StaticCyclicPollScheduler(PollScheduler):
    def __init__(self, *args, **kwargs):
        super(StaticCyclicPollScheduler, self).__init__(*args, **kwargs)
        # Slot Plans has: {remote: {hyperperiod: {slot: WeakSet(points)}}}
        self.slot_plans: list[dict[timedelta, dict[timedelta, list[PollSet]]]] = []

    def get_schedule(self, full_topics=False):
        """Return the calculated schedules to the user."""
        return_dict = defaultdict(lambda: defaultdict(dict))
        for slot_plan in self.slot_plans:
            for hyperperiod, plan in slot_plan.items():
                for slot, poll_sets in plan.items():
                    poll_set = reduce(lambda ps1, ps2: ps1 | ps2, poll_sets)
                    remote = str(poll_set.remote.unique_id)
                    if full_topics:
                        return_dict[str(hyperperiod)][str(slot)][remote] = [p for p in poll_set.points.keys()]
                    else:
                        return_dict[str(hyperperiod)][str(slot)][remote] = [p.split("/")[-1] for p in poll_set.points.keys()]
        return return_dict

    @staticmethod
    def calculate_hyperperiod(intervals, minimum_polling_interval):
        return lcm(*[floor(i / minimum_polling_interval) for i in intervals]) * minimum_polling_interval

    @staticmethod
    def _separate_coprimes(intervals):
        # TODO: The math in _separate_coprimes and calculate_hyperperiod needs to be done with integers, not floats.
        #  How to handle sub-second intervals?  Should we do this using milliseconds?
        separated = []
        unseparated = list(intervals)
        unseparated.sort(reverse=True)
        while len(unseparated) > 0:
            non_coprime, coprime = [], []
            first = unseparated.pop(0)
            non_coprime.append(first)
            for i in unseparated:
                if gcd(first, i) == 1 and first != 1 and i != 1:
                    coprime.append(i)
                else:
                    non_coprime.append(i)
            unseparated = coprime
            separated.append(non_coprime)
        return separated

    def _find_slots(self, input_dict: dict[float, dict[DriverAgent, list[PollSet]]], parallel_remote_index: int = 0):
        coprime_interval_sets = self._separate_coprimes(input_dict.keys())
        slot_plan: dict[timedelta, dict[timedelta, list[PollSet]]] = defaultdict(lambda: defaultdict(list))
        parallel_offset = parallel_remote_index * self.group_config.minimum_polling_interval
        min_spread = self.group_config.minimum_polling_interval
        all_remotes = {k for i in input_dict for k in input_dict[i].keys()}
        min_interval = min(input_dict.keys())
        min_remote_offset = min_interval / len(all_remotes)
        if self.group_config.parallel_subgroups and min_remote_offset < min_spread:
            _log.warning(f'There are {len(all_remotes)} scheduled sequentially with a smallest interval of'
                         f' {min_interval}. This only allows {min_remote_offset} between polls --- less than'
                         f' the group {self.group} minimum_polling_interval of {min_spread}. The resulting schedule is'
                         f' likely to result in unexpected behavior and potential loss of data if these remotes share'
                         f' a collision domain. If the minimum polling interval cannot be lowered, consider polling'
                         f' less frequently.')
        remote_offsets = {r: i * min_remote_offset for i, r in enumerate(all_remotes)}
        for interval_set in coprime_interval_sets:
            hyperperiod = self.calculate_hyperperiod(interval_set, min(interval_set))
            for interval in interval_set:
                s_count = int(hyperperiod / interval)
                remote_spread = interval / len(input_dict[interval].keys())
                spread = min_spread if self.group_config.parallel_subgroups else max(min_spread, remote_spread)
                for slot, remote in [((interval * i) + (spread * r) + remote_offsets[remote] + parallel_offset , remote)
                                     for i in range(s_count) for r, remote in enumerate(input_dict[interval].keys())]:
                    slot_plan[timedelta(seconds=hyperperiod)][timedelta(seconds=slot)].extend(input_dict[interval][remote])
        # noinspection PyTypeChecker
        return {hyperperiod: dict(sorted(sp.items())) for hyperperiod, sp in slot_plan.items()}

    @staticmethod
    def get_poll_generator(hyperperiod_start: datetime, hyperperiod: timedelta, slot_plan: dict[timedelta, list[PollSet]]):
        def get_polls(start_time):
            # Union of points and publish_setups is here to get any changes to self.poll_sets at start of hyperperiod.
            return ((start_time + slot, reduce(lambda d1, d2: d1 | d2, poll_sets)
                     ) for slot, poll_sets in slot_plan.items())
        polls = get_polls(hyperperiod_start)
        while True:
            try:
                p = next(polls)
            except StopIteration:
                hyperperiod_start += hyperperiod
                polls = get_polls(hyperperiod_start)
                p = next(polls)
            yield p

    def _prepare_to_schedule(self):
        group_poll_sets = self.poll_sets[self.group]
        if self.group_config.parallel_subgroups:
            for parallel_index, (remote, remote_poll_sets) in enumerate(group_poll_sets.items()):
                input_dict = defaultdict(lambda: defaultdict(list))
                for interval, poll_set in remote_poll_sets.items():
                    input_dict[interval][remote].append(poll_set)
                self.slot_plans.append(self._find_slots(input_dict, parallel_index))
        else:
            input_dict = defaultdict(lambda: defaultdict(list))
            for remote, remote_poll_sets in group_poll_sets.items():
                for interval, poll_set in remote_poll_sets.items():
                    input_dict[interval][remote].append(poll_set)
            self.slot_plans.append(self._find_slots(input_dict))

    def _schedule_polling(self):
        # TODO: How to fully ensure min_polling_interval? Nothing yet prevents collisions between individual polls in
        #  separate schedules. Is it worth keeping these apart if it requires a check for each slot at schedule time?
        #  Or, create global lock oscillating at min_poll_interval - check on poll for the next allowed start time?
        _log.debug('In _schedule_polling, poll sets is: ')
        _log.debug(self.slot_plans)
        for slot_plan in self.slot_plans:
            for hyperperiod, plan in slot_plan.items():
                initial_start = self.find_starting_datetime(get_aware_utc_now(), hyperperiod,
                                                            self.group_config.start_offset)
                self.start_all_datetime = max(self.start_all_datetime, initial_start + hyperperiod)
                poll_generator = self.get_poll_generator(initial_start, hyperperiod, plan)
                start, poll_set = next(poll_generator)
                _log.info(f'Scheduled polling for {self.group}--{hyperperiod} starts at {start.time()} (datetime: {start})')
                # TODO: Is hyperperiod a sufficient index for the pollers?
                self.pollers[hyperperiod] = self.data_model.agent.core.schedule(start, self._operate_polling,
                                                                                hyperperiod, poll_generator, poll_set)

    def _operate_polling(self, poller_id, poll_generator, current_poll_set):
        next_start, next_poll_set = next(poll_generator)

        # Find the current and next polls where the next poll is the first to still be in the future
        #  (This assures that if the host has gone to sleep, the poll will still be the most up to date):
        now = get_aware_utc_now()
        while next_start <= now:
            # TODO: If this takes too long for long pauses, call get_poll_generator again, instead.
            _log.warning(f'Skipping polls from {next_start} to {now} to catch up to the current time.')
            current_poll_set = next_poll_set
            next_start, next_poll_set = next(poll_generator)

        # Schedule next poll:
        if next_poll_set.points:
            self.pollers[poller_id] = self.data_model.agent.core.schedule(next_start, self._operate_polling, poller_id,
                                                                          poll_generator, next_poll_set)
        else:
            _log.info(f'Stopping polling loop of {poller_id} points on {next_poll_set.remote.unique_id}.'
                      f' There are no points in this request set to poll.')
        current_poll_set.remote.poll_data(current_poll_set)


class SerialPollScheduler(PollScheduler):
    def get_schedule(self):
        pass

    def _prepare_to_schedule(self):
        pass

    def __init__(self, agent, sleep_duration, **kwargs):
        super(SerialPollScheduler, self).__init__(agent, **kwargs)
        self.sleep_duration = sleep_duration

        self.status = {}

    # TODO: Serial Poll Scheduler (schedule a single job to poll each item of poll set after the return or failure
    #  of the previous):
    #  Create timeouts such that enough time is available to address each item in series before the next cycle.
    def _schedule_polling(self):
        pass

    # TODO: If there is not sufficient difference in the prepare and schedule methods,
    #  this could be a separate operate method in the StaticCyclicPollScheduler.
    def _operate_polling(self,  remote, poller_id, poll_generator):
        while True:  # TODO: This should probably check whether agent is stopping.
            start, points = next(poll_generator)
            poll = gevent.spawn(remote.poll_data, points)
            # Wait for poll to finish.
            while not poll.ready():
                gevent.sleep(self.sleep_duration)
            # Track whether this poller_id has been successful.
            # TODO: Would it be more helpful if the poll_data method returned the time (when it is successful) or None?
            self.status[poller_id] = poll.get(timeout=1)
